# -*- coding: utf-8 -*-


from datetime import datetime,timedelta

def human_readable_bytes(value, digits= 2, delim= "", postfix=""):
    """Return a human-readable file size.
    """
    if value is None:
        return None
    chosen_unit = "B"
    for unit in ("KiB", "MiB", "GiB", "TiB"):
        if value > 1000:
            value /= 1024
            chosen_unit = unit
        else:
            break
    return f"{value:.{digits}f}" + delim + chosen_unit + postfix

def human_readable_timedelta(seconds, precision = 0):
    """Return a human-readable time delta as a string.
    """
    pieces = []
    value = timedelta(seconds=seconds)
    

    if value.days:
        pieces.append(f"{value.days}d")

    seconds = value.seconds

    if seconds >= 3600:
        hours = int(seconds / 3600)
        pieces.append(f"{hours}h")
        seconds -= hours * 3600

    if seconds >= 60:
        minutes = int(seconds / 60)
        pieces.append(f"{minutes}m")
        seconds -= minutes * 60

    if seconds > 0 or not pieces:
        pieces.append(f"{seconds}s")

    if not precision:
        return "".join(pieces)

    return "".join(pieces[:precision])



# -*- coding: utf-8 -*-


from telethon import TelegramClient
import asyncio

class tgtkclient(TelegramClient):
    def __init__(self,*args,**kwargs):
        super().__init__(*args,**kwargs)
        
        #for now only queue is required
        self.queue = None
        self.dl_passwords = {}
        self.pyro = None


# -*- coding: utf-8 -*-


from ..functions import vids_helpers
from hachoir.parser import createParser
from hachoir.metadata import extractMetadata
import random

async def get_thumbnail(file_path, user_id = None):
    metadata = extractMetadata(createParser(file_path))
    try:
        duration = metadata.get("duration")
    except:
        duration = 3

    if user_id is not None:
        pass # todo code for custom thumbnails here mostly will be with db
    else:
        path = await vids_helpers.gen_ss(file_path,random.randint(2,duration.seconds))
        path = await vids_helpers.resize_img(path,320)
        return path


# -*- coding: utf-8 -*-


from ..consts.DefaultCommands import Commands
from ..core.getVars import get_val 
import os,logging

torlog = logging.getLogger(__name__)

def get_command(command):
    cmd = None

    #Get the command from the constants supplied
    try:
        cmd = getattr(Commands,command)
        torlog.debug(f"getting the command {command} from file: {cmd}")
    except AttributeError:pass

    #Get the commands form the env [overlap]
    #try:
    envcmd = os.environ.get(command)
    torlog.debug(f"hetting the command {command} from file: {envcmd}")
    cmd =  envcmd if envcmd is not None else cmd

    #Get the commands form the DB [overlap]
    #TODO database

    if cmd is None:
        torlog.debug(f"None Command Error occured for command {command}")
        raise Exception("the command was not found in either the constants, environment or database. command is: {}".format(command))
    
    cmd = cmd.strip("/")
    cmd += get_val("BOT_CMD_POSTFIX")

    torlog.debug(f"final resolver for {command} is {cmd}")
    return f"/{cmd}"


from speedtest import Speedtest
import logging
from ..functions.Human_Format import human_readable_bytes

torlog = logging.getLogger(__name__)

async def get_speed(message):
    imspd = await message.reply("`Running speedtest...`")
    test = Speedtest()
    test.get_best_server()
    test.download()
    test.upload()
    test.results.share()
    result = test.results.dict()
    path = (result['share'])
    string_speed = f'''
**Speedtest Result:-**
Server Name: `{result["server"]["name"]}`
Country: `{result["server"]["country"]}, {result["server"]["cc"]}`
Sponsor: `{result["server"]["sponsor"]}`
Upload: `{human_readable_bytes(result["upload"] / 8)}/s`
Download: `{human_readable_bytes(result["download"] / 8)}/s`
Ping: `{result["ping"]} ms`
ISP: `{result["client"]["isp"]}`
'''
    await imspd.delete()
    await message.reply(string_speed, parse_mode="markdown")
    torlog.info(f'Server Speed result:-\nDL: {human_readable_bytes(result["download"] / 8)}/s UL: {human_readable_bytes(result["upload"] / 8)}/s')


# -*- coding: utf-8 -*-


from .status import Status, QBTask
from ..getVars import get_val
import os, re, logging
from telethon.errors.rpcerrorlist import MessageNotModifiedError, FloodWaitError

torlog = logging.getLogger(__name__)


class TGUploadTask(Status):
    
    def __init__(self, task):
        super().__init__()
        self.Tasks.append(self)
        self._dl_task = task
        self._files = 0
        self._dirs = 0
        self._uploaded_files = 0
        self._active = True
        self._current_file = ""
        self._message = None
        self._omess = None
        self.cancel = False

    async def get_message(self):
        return self._message

    async def get_sender_id(self):
        return self._omess.sender_id

    async def get_original_message(self):
        return self._omess

    async def set_message(self, message):
        self._message = message
    
    async def set_original_message(self, omess):
        self._omess = omess

    async def set_inactive(self):
        self._active = False

    async def is_active(self):
        return self._active

    async def add_a_dir(self, path):
        await self.dl_files(path)

    async def dl_files(self, path = None):
        if path is None:
            path = await self._dl_task.get_path()
        
        if os.path.isfile(path):
            self._files += 1
            return
        
        files = self._files
        dirs = self._dirs
        for _, d, f in os.walk(path, topdown=False):
            for _ in f:
                files += 1
            for _ in d:
                dirs += 1
        
        # maybe will add blacklisting of Extensions
        self._files = files
        self._dirs = dirs

    async def uploaded_file(self, name=None):
        self._uploaded_files += 1
        print("\n----updates files to {}\n".format(self._uploaded_files))
        self._current_file = str(name)

    async def create_message(self):
        msg = "<b>uploading:- </b> <code>{}</code>\n".format(
            self._current_file
        )
        prg = 0
        try:
            prg = self._uploaded_files/self._files

        except ZeroDivisionError:pass
        msg += "<b>progress: </b> {} - {}%\n".format(
            self.progress_bar(prg),
            prg*100
        )
        msg += "<b>files: </b> {} of {} done.\n".format(
            self._uploaded_files,
            self._files
        )
        msg += "<b>using engine: </b> <code>telegram upload</code>\n"
        return msg

    def progress_bar(self, percentage):
        """Returns a progress bar for download
        """
        #percentage is on the scale of 0-1
        comp = get_val("COMPLETED_STR")
        ncomp = get_val("REMAINING_STR")
        pr = ""

        for i in range(1,11):
            if i <= int(percentage*10):
                pr += comp
            else:
                pr += ncomp
        return pr

class RCUploadTask(Status):
    def __init__(self, task):
        super().__init__()
        self.Tasks.append(self)
        self._dl_task = task
        self._active = True
        self._upmsg = ""
        self._prev_cont = ""
        self._message = None
        self._error = ""
        self._omess = None
        self.cancel = False

    async def set_original_message(self, omess):
        self._omess = omess

    async def get_original_message(self):
        return self._omess
    
    async def get_sender_id(self):
        return self._omess.sender_id

    async def set_message(self, message):
        self._message = message
    
    async def refresh_info(self,msg):
        # The rclone is process dependent so cant be updated here.
        self._upmsg = msg

    async def create_message(self):
        mat = re.findall("transferred:.*eta.*",self._upmsg)
        nstr = mat[0].replace("transferred:","")
        nstr = nstr.strip()
        nstr = nstr.split(",")
        prg = nstr[1].strip("% ")
        prg = "progress: {} - {}%".format(self.progress_bar(prg),prg)
        progress = "<b>uploaded: {} \n{} \nspeed: {} \neta: {}</b> \n<b>using engine: </b><code>rclone</code>".format(nstr[0],prg,nstr[2],nstr[3].replace("eta",""))
        return progress

    def progress_bar(self,percentage):
        """Returns a progress bar for download
        """
        #percentage is on the scale of 0-1
        comp = get_val("COMPLETED_STR")
        ncomp = get_val("REMAINING_STR")
        pr = ""

        try:
            percentage=int(percentage)
        except:
            percentage = 0

        for i in range(1,11):
            if i <= int(percentage/10):
                pr += comp
            else:
                pr += ncomp
        return pr

    async def update_message(self):
        progress = await self.create_message()
        if not self._prev_cont == progress:
            #kept just in case
            self._prev_cont = progress
            try:
                await self._message.edit(progress,parse_mode="html")
            except MessageNotModifiedError as e:
                torlog.debug("{}".format(e))
            except FloodWaitError as e:
                torlog.error("{}".format(e))
            except Exception as e:
                torlog.info("Not expected {}".format(e))

    async def is_active(self):
        return self._active

    async def set_inactive(self, error = None):
        self._active = False
        if error is not None:
            self._error = error


# -*- coding: utf-8 -*-


from telethon.tl.functions.channels import GetParticipantRequest
from telethon.tl.types import ChannelParticipantAdmin,ChannelParticipantCreator,ChannelParticipantsAdmins
import logging,traceback
from ..core.getVars import get_val
torlog = logging.getLogger(__name__)

#todo add alpha admin if needed

async def is_admin(client,user_id,chat_id, force_owner=False):
    if force_owner:
        if user_id == get_val("OWNER_ID"):
            return True
        else:
            return False
        
    try:
        res = await client(GetParticipantRequest(
            channel=chat_id,
            user_id=user_id
        ))

        try:
            if isinstance(res.participant,(ChannelParticipantAdmin,ChannelParticipantCreator,ChannelParticipantsAdmins)):
                return True
            else:
                
                if user_id in get_val("ALD_USR"):
                    return True
                else:
                    return False
        except:
            torlog.info("Bot Accessed in Private {}".format(traceback.format_exc()))
            return False
    except Exception as e:
        torlog.info("Bot Accessed in Private {}".format(e))
        if user_id in get_val("ALD_USR"):
            return True
        else:
            return False


#!/usr/bin/env python
# -*- coding: utf-8 -*-
from telethon import TelegramClient
from tgtk.core.HandleManager import add_handlers
from tgtk.core.getVars import get_val
import logging,asyncio
from tgtk.core.wserver import start_server_async
from tgtk.core.status.auto_delete import del_status
from pyrogram import Client
try:
    from tgtk.functions.rstuff import get_rstuff
except ImportError:pass

from tgtk.ttk_client import tgtkclient

if __name__ == "__main__":

    #logging stuff
    #thread name is just kept for future use
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s %(levelname)s %(threadName)s %(name)s %(message)s"
    )
    logging.getLogger("pyrogram").setLevel(logging.ERROR)
    
    # parallel connections limiter
    queue = asyncio.Queue()
    exqueue = asyncio.Queue()
    
    for i in range(1,4):
        queue.put_nowait(i)

    for i in range(1,5):
        exqueue.put_nowait(i)
    
    # Telethon client creation
    ttkbot = tgtkclient("tgtkbot",get_val("API_ID"),get_val("API_HASH"), timeout=20, retry_delay=3, request_retries=10, connection_retries=10)
    ttkbot.queue = queue
    ttkbot.exqueue = exqueue
    ttkbot.start(bot_token=get_val("BOT_TOKEN"))
    logging.info("telethon client created.")

    # Pyro Client creation and linking
    pyroclient = Client("pyrosession", api_id=get_val("API_ID"), api_hash=get_val("API_HASH"), bot_token=get_val("BOT_TOKEN"), workers=100)
    pyroclient.start()
    ttkbot.pyro = pyroclient
    logging.info("pyrogram client created.")

    # Associate the handlers
    add_handlers(ttkbot)

    ttkbot.loop.create_task(del_status())

    if get_val("IS_VPS"):
        ttkbot.loop.run_until_complete(start_server_async(get_val("SERVPORT")))
    try:
        ttkbot.loop.run_until_complete(get_rstuff())
    except:pass
    
    logging.info("it's on!")

    ttkbot.run_until_disconnected()


# -*- coding: utf-8 -*-

import logging
from ...functions.Human_Format import human_readable_bytes, human_readable_timedelta
from ..getVars import get_val
from telethon.errors.rpcerrorlist import MessageNotModifiedError, FloodWaitError
from telethon.tl.types import KeyboardButtonCallback
from datetime import datetime

torlog = logging.getLogger(__name__)

class Status():
    # Shared List
    Tasks = []
    
    def __init__(self):
        self._task_id = len(self.Tasks) + 1

    def refresh_info(self):
        raise NotImplementedError

    def update_message(self):
        raise NotImplementedError

    def is_active(self):
        raise NotImplementedError

    def set_inactive(self):
        raise NotImplementedError

# qBittorrent Task Class
class QBTask(Status):
    
    def __init__(self, torrent, message, client):
        super().__init__()
        self.Tasks.append(self)
        self.hash = torrent.hash
        self._torrent = torrent
        self._message = message
        self._client = client
        self._active = True
        self._path = torrent.save_path
        self._error = ""
        self._done = False
        self.cancel = False
        self._omess = None
        self._prevmsg = ""
    
    async def set_original_mess(self, omess):
        self._omess = omess
    
    async def get_original_message(self):
        return self._omess

    async def refresh_info(self, torrent = None):
        if torrent is None:
            self._torrent = self._client.torrents_info(torrent_hashes=self._torrent.hash)
        else:
            self._torrent = torrent

    async def get_sender_id(self):
        return self._omess.sender_id

    async def create_message(self):
        msg = "<b>downloading:</b> <code>{}</code>\n".format(
            self._torrent.name
            )
        msg += "<b>down:</b> {} <b>up:</b> {}\n".format(
            human_readable_bytes(self._torrent.dlspeed,postfix="/s"),
            human_readable_bytes(self._torrent.upspeed,postfix="/s")
            )
        msg += "<b>progress:</b> {} - {}%\n".format(
            self.progress_bar(self._torrent.progress),
            round(self._torrent.progress*100,2)
            )
        msg += "<b>downloaded:</b> {} out of {}\n".format(
            human_readable_bytes(self._torrent.downloaded),
            human_readable_bytes(self._torrent.total_size)
            )
        msg += "<b>eta:</b> <b>{}</b>\n".format(
            human_readable_timedelta(self._torrent.eta)
            )
        msg += "<b>s:</b>{} <b>l:</b>{}\n".format(
            self._torrent.num_seeds,self._torrent.num_leechs
            )
        msg += "<b>using engine:</b> <code>qBittorrent</code>"

        return msg

    async def get_state(self):
        #stalled
        if self._torrent.state == "stalledDL":
            return"torrent <code>{}</code> stalled (waiting for connection) temporarily.".format(self._torrent.name)
        #meta stage
        elif self._torrent.state == "metaDL":
            return  "getting metadata for {} - {}".format(self._torrent.name,datetime.now().strftime("%H:%M:%S"))
        elif self._torrent.state == "downloading" or self._torrent.state.lower().endswith("dl"):
            # kept for past ref
            return None

    async def central_message(self):
        cstate = await self.get_state()
        if cstate is not None:
            return cstate
        else:
            return await self.create_message()

    async def update_message(self):
        msg = await self.create_message()
        if self._prevmsg == msg:
            return

        self._prevmsg = msg
        
        try:
        
            cstate = await self.get_state()
            
            msg = cstate if cstate is not None else msg
            
            await self._message.edit(msg,parse_mode="html",buttons=self._message.reply_markup) 

        except MessageNotModifiedError as e:
            torlog.debug("{}".format(e))
        except FloodWaitError as e:
            torlog.error("{}".format(e))
        except Exception as e:
            torlog.info("not expected {}".format(e))


    async def set_done(self):
        self._done = True
        await self.set_inactive()

    def is_done(self):
        return self._done

    async def set_path(self, path):
        self._path = path

    async def get_path(self):
        return self._path

    async def set_inactive(self, error=None):
        self._active = False
        if error is not None:
            self._error = error

    async def is_active(self):
        return self._active

    def progress_bar(self, percentage):
        """Returns a progress bar for download
        """
        #percentage is on the scale of 0-1
        comp = get_val("COMPLETED_STR")
        ncomp = get_val("REMAINING_STR")
        pr = ""

        for i in range(1,11):
            if i <= int(percentage*10):
                pr += comp
            else:
                pr += ncomp
        return pr


class ARTask(Status):
    
    def __init__(self, gid, message, aria2, dl_file):
        super().__init__()
        self.Tasks.append(self)
        self._gid = gid
        self._dl_file = dl_file 
        self._message = message
        self._aria2 = aria2
        self._active = True
        self._error = ""
        self._done = False
        self.cancel = False
        self._omess = None
        self._path =None 
        self._prevmsg = ""
    # Setters

    async def set_original_mess(self, omess=None):
        if omess is None:
            omess = await self._message.get_reply_message()

        self._omess = omess

    async def get_original_message(self):
        return self._omess

    async def get_gid(self):
        return self._gid

    async def set_gid(self, gid):
        self._gid = gid
    
    async def get_sender_id(self):
        return self._omess.sender_id

    async def refresh_info(self, dl_file = None):
        if dl_file is None:
            try:
                self._dl_file = self._aria2.get_download(self._gid)
            except:
                torlog.exception("error fetching direct download link.")
        else:
            self._dl_file = dl_file

    async def create_message(self):
        # Getting the vars pre handed
        downloading_dir_name = "N/A"
        try:
            downloading_dir_name = str(self._dl_file.name)
        except:
            pass

        msg = "<b>downloading:</b> <code>{}</code>\n".format(
            downloading_dir_name
            )
        msg += "<b>down:</b> {} <b>Up:</b> {}\n".format(
            self._dl_file.download_speed_string(),
            self._dl_file.upload_speed_string()
            )
        msg += "<b>progress:</b> {} - {}%\n".format(
            self.progress_bar(self._dl_file.progress/100),
            round(self._dl_file.progress,2)
            )
        msg += "<b>downloaded:</b> {} out of {}\n".format(
            human_readable_bytes(self._dl_file.completed_length),
            human_readable_bytes(self._dl_file.total_length)
            )
        msg += "<b>eta:</b> <b>{}</b>\n".format(
            self._dl_file.eta_string()
            )
        msg += "<b>conns:  </b>{} <b>\n".format(
            self._dl_file.connections
            )
        msg += "<b>using engine:</b> <code>aria2</code>"

        return msg

    async def get_state(self):
        # No states for aria2
        pass

    async def central_message(self):
        return await self.create_message()

    async def update_message(self):
        msg = await self.create_message()
        if self._prevmsg == msg:
            return
        
        self._prevmsg = msg
        
        try:
            data = "torcancel aria2 {} {}".format(
                self._gid,
                self._omess.sender_id
            )
            await self._message.edit(msg,parse_mode="html",buttons=[KeyboardButtonCallback("cancel direct leech",data=data.encode("UTF-8"))]) 

        except MessageNotModifiedError as e:
            torlog.debug("{}".format(e))
        except FloodWaitError as e:
            torlog.error("{}".format(e))
        except Exception as e:
            torlog.info("Not expected {}".format(e))

    async def set_done(self):
        self._done = True
        await self.set_inactive()

    def is_done(self):
        return self._done

    async def set_inactive(self, error=None):
        self._active = False
        if error is not None:
            self._error = error

    async def is_active(self):
        return self._active

    async def get_error(self):
        return self._error

    async def set_path(self, path):
        self._path = path

    async def get_path(self):
        return self._path

    def progress_bar(self, percentage):
        """Returns a progress bar for download
        """
        #percentage is on the scale of 0-1
        comp = get_val("COMPLETED_STR")
        ncomp = get_val("REMAINING_STR")
        pr = ""

        for i in range(1,11):
            if i <= int(percentage*10):
                pr += comp
            else:
                pr += ncomp
        return pr


# -*- coding: utf-8 -*-
# (c) YashDK [yash-dk@github]

from telethon.tl.types import KeyboardButtonCallback,KeyboardButton
from telethon import events
from tgtk import SessionVars
import asyncio as aio
from .getVars import get_val
from .database_handle import tkdb
from .. import user_db
from functools import partial
import time,os,configparser,logging,traceback
from PIL import Image

torlog = logging.getLogger(__name__)
#logging.getLogger("telethon").setLevel(logging.DEBUG)

TIMEOUT_SEC = 60

# this file will contian all the handlers and code for settings
# code can be more modular i think but not bothering now
# todo make the code more modular
no = "✖"
yes = "✔"
# Central object is not used its Acknowledged 
tordb = tkdb()
header =  '<b>tgtk - a telegram leecher bot.\n</b>\n<u>user settings menu</u>'
async def handle_user_setting_callback(e):
    db = tordb
    sender_id = str(e.sender_id)

    
    data = e.data.decode()
    cmd = data.split(" ")
    
    val = ""
    
    if cmd[-1] != sender_id:
        print("Sender id",sender_id," - - ",cmd[-1])
        await e.answer("not yours.",alert=True)
        #await e.delete()
        return
    if cmd[1] == "mycmd":
        pass
    elif cmd[1] == "rclonemenu":
        # this is menu
        mmes = await e.get_message()
        await handle_user_settings(mmes,True,"\nWelcome to Rclone Config Menu. TD= Team Drive, ND= Normal Drive",submenu="rclonemenu",sender_id=sender_id)
    elif cmd[1] == "thumbmenu":
        # this is menu
        mmes = await e.get_message()
        await handle_user_settings(mmes,True,"\nthumbnail menu.",submenu="thumbmenu",sender_id=sender_id)
    elif cmd[1] == "rcloneconfig":
        await e.answer("send the rclone config file which you have generated.",alert=True)
        mmes = await e.get_message()
        await mmes.edit(f"{mmes.raw_text}\n/ignore to go back",buttons=None)
        val = await get_value(e,True)
        
        await general_input_manager(e,mmes,"RCLONE_CONFIG","str",val,sender_id,"rclonemenu")  
    elif cmd[1] == "setthumb":
        await e.answer("send the thumbnail.", alert=True)
        mmes = await e.get_message()
        await mmes.edit(f"{mmes.raw_text}\n /ignore to go back.", buttons=None)
        val = await get_value(e,file=True,photo=True)
        await general_input_manager(e,mmes,"THUMBNAIL","str",val,sender_id,"thumbmenu")  

    elif cmd[1] == "selfdest":
        await e.answer("closed.")
        await e.delete()
    elif cmd[1] == "change_drive":
        await e.answer(f"changed default drive to {cmd[2]}.",alert=True)
        user_db.set_var("DEF_RCLONE_DRIVE",str(cmd[2]), e.sender_id)

        await handle_user_settings(await e.get_message(),True,f"<b><u>changed the default drive to {cmd[2]}</b></u>","rclonemenu",sender_id=sender_id)
    elif cmd[1] == "mainmenu":
        # this is menu
        mmes = await e.get_message()
        await handle_user_settings(mmes,True,sender_id=sender_id)
    elif cmd[1] == "fdocs":
        await e.answer("")
        if cmd[2] == "true":
            val = True
        else:
            val = False
        
        user_db.set_var("FORCE_DOCUMENTS",val, str(e.sender_id))
        await handle_user_settings(await e.get_message(),True,f"<b><u>changed the value to {val} of force documents.</b></u>",sender_id=sender_id)
    elif cmd[1] == "disablethumb":
        await e.answer("")
        if cmd[2] == "true":
            val = True
        else:
            val = False
        
        user_db.set_var("DISABLE_THUMBNAIL",val, str(e.sender_id))
        await handle_user_settings(await e.get_message(),True,f"<b><u>changed the value to {val} of disable thumbnail.</b></u>",sender_id=sender_id, submenu="thumbmenu")



async def handle_user_settings(e,edit=False,msg="",submenu=None,sender_id=None):
    # this function creates the menu
    # and now submenus too
    if sender_id is None:
        sender_id=e.sender_id
    session_id=None
    
    menu = [
        #[KeyboardButtonCallback(yes+" Allow TG Files Leech123456789-","settings data".encode("UTF-8"))], # for ref
    ]
    
    if submenu is None:
        await get_bool_variable("FORCE_DOCUMENTS","FORCE_DOCUMENTS",menu,"fdocs",sender_id)#
        #await get_string_variable("RCLONE_CONFIG",menu,"rcloneconfig",session_id)
        await get_sub_menu("open rclone menu","rclonemenu",sender_id,menu)#
        await get_sub_menu("open thumbnail menu","thumbmenu",sender_id,menu)#
        # thumbnail
        menu.append(
            [KeyboardButtonCallback("close menu.",f"usettings selfdest {sender_id}".encode("UTF-8"))]
        )


        if edit:
            rmess = await e.edit(header+"\nEnjoy.\n"+msg,parse_mode="html",buttons=menu,link_preview=False, file="tk.jpg")
        else:
            rmess = await e.reply(header+"\nEnjoy.\n",parse_mode="html",buttons=menu,link_preview=False, file="tk.jpg")
    elif submenu == "rclonemenu":
        rcval = await get_string_variable("RCLONE_CONFIG",menu,"rcloneconfig",sender_id)
        if rcval != "None":
            # create a all drives menu
            if not "not loaded" in rcval:
                
                path = user_db.get_rclone(sender_id)
                
                conf = configparser.ConfigParser()
                conf.read(path)
                #menu.append([KeyboardButton("Choose a default drive from below")])
                def_drive = user_db.get_var("DEF_RCLONE_DRIVE", sender_id)

                for j in conf.sections():
                    prev=""
                    if j == def_drive:
                        prev = yes

                    if "team_drive" in list(conf[j]):
                        menu.append(
                            [KeyboardButtonCallback(f"{prev}{j} - TD",f"usettings change_drive {j} {sender_id}")]
                        )
                    else:
                        menu.append(
                            [KeyboardButtonCallback(f"{prev}{j} - ND",f"usettings change_drive {j} {sender_id}")]
                        )

        await get_sub_menu("go back ⬅️","mainmenu",sender_id,menu)
        menu.append(
            [KeyboardButtonCallback("close menu",f"usettings selfdest {sender_id}".encode("UTF-8"))]
        )
        if edit:
            rmess = await e.edit(header+"\nit is recommended to lock the group before setting vars.\n"+msg,parse_mode="html",buttons=menu,link_preview=False)


    elif submenu == "thumbmenu":
        thumb = user_db.get_thumbnail(sender_id)
        if thumb is not False:
            menu.append(
                [KeyboardButtonCallback("change thumbnail", f"usettings setthumb {sender_id}".encode("UTF-8"))]
            )
            await get_bool_variable("DISABLE_THUMBNAIL","Disable Thumbnail",menu,"disablethumb",sender_id)
            await get_sub_menu("go back ⬅️","mainmenu",sender_id,menu)
            menu.append(
                [KeyboardButtonCallback("close menu",f"usettings selfdest {sender_id}".encode("UTF-8"))]
            )
            await e.edit(header+"\nmanage your thumbnail(s) on the fly.", file=thumb, buttons=menu, parse_mode="html")
        else:
            menu.append(
                [KeyboardButtonCallback("set thumbnail.", f"usettings setthumb {sender_id}".encode("UTF-8"))]
            )
            await get_sub_menu("go back ⬅️","mainmenu",sender_id,menu)
            menu.append(
                [KeyboardButtonCallback("close menu",f"usettings selfdest {sender_id}".encode("UTF-8"))]
            )
            await e.edit(header+"\nmanage your thumbnail(s) on the fly.", parse_mode="html", buttons=menu)

# an attempt to manager all the input
async def general_input_manager(e,mmes,var_name,datatype,value,sender_id,sub_menu):
    if value is not None and not "ignore" in value:
        await confirm_buttons(mmes,value)
        conf = await get_confirm(e)
        if conf is not None:
            if conf:
                try:
                    if datatype == "int":
                        value = int(value)
                    if datatype == "str":
                        value = str(value)
                    if datatype == "bool":
                        if value.lower() == "true":
                            value = True
                        elif value.lower() == "false":
                            value = False
                        else:
                            raise ValueError("Invalid value from bool")
                    
                    if var_name == "RCLONE_CONFIG":
                        #adjust the special case
                        try:
                            conf = configparser.ConfigParser()
                            conf.read(value)
                            for i in conf.sections():
                                user_db.set_var("DEF_RCLONE_DRIVE",str(i), e.sender_id)
                                break
                                
                            with open(value,"rb") as fi:
                                data = fi.read()
                                user_db.set_rclone(data, e.sender_id)
                            os.remove(value)
                            #db.set_variable("LEECH_ENABLED",True)
                            #SessionVars.update_var("LEECH_ENABLED",True)
                        except Exception:
                            torlog.error(traceback.format_exc())
                            await handle_user_settings(mmes,True,f"<b><u>the configuration file is invalid, check logs.</b></u>",sub_menu)
                            return
                    elif var_name == "THUMBNAIL":
                        try:
                            im = Image.open(value)
                            im.convert("RGB").save(value,"JPEG")
                            im = Image.open(value)
                            im.thumbnail((320,320), Image.ANTIALIAS)
                            im.save(value,"JPEG")
                            with open(value,"rb") as fi:
                                data = fi.read()
                                user_db.set_thumbnail(data, e.sender_id)
                            os.remove(value)
                        except Exception:
                            torlog.error(traceback.format_exc())
                            await handle_user_settings(mmes,True,f"<b><u>an error occured with the thumbnail you sent.</b></u>",sub_menu)
                            return
                    else:
                        user_db.set_var(var_name, value, e.sender_id)
                        #db.set_variable(var_name,value)
                        #SessionVars.update_var(var_name,value)
                    
                    await handle_user_settings(mmes,True,f"<b><u>received {var_name} value '{value}' with confirm.</b></u>",sub_menu, sender_id=sender_id)
                except ValueError:
                    await handle_user_settings(mmes,True,f"<b><u>value [{value}] not valid try again and enter {datatype}.</b></u>",sub_menu, sender_id=sender_id)    
            else:
                await handle_user_settings(mmes,True,f"<b><u>confirm differed by user.</b></u>",sub_menu, sender_id=sender_id)
        else:
            await handle_user_settings(mmes,True,f"<b><u>confirm timed out [waited 60s for input].</b></u>",sub_menu, sender_id=sender_id)
    else:
        await handle_user_settings(mmes,True,f"<b><u>entry timed out [waited 60s for input].</b></u>",sub_menu, sender_id=sender_id)


async def get_value(e,file=False,photo=False):
    # todo replace with conver. - or maybe not Fix Dont switch to conversion
    # this function gets the new value to be set from the user in current context
    lis = [False,None]

    #func tools works as expected ;);)    
    cbak = partial(val_input_callback,o_sender=e.sender_id,lis=lis,file=file,photo=photo)
    
    e.client.add_event_handler(
        #lambda e: test_callback(e,lis),
        cbak,
        events.NewMessage()
    )

    start = time.time()

    while not lis[0]:
        if (time.time() - start) >= TIMEOUT_SEC:
            break

        await aio.sleep(1)
    
    val = lis[1]
    
    e.client.remove_event_handler(cbak)

    return val

async def get_confirm(e):
    # abstract for getting the confirm in a context

    lis = [False,None]
    cbak = partial(get_confirm_callback,o_sender=e.sender_id,lis=lis)
    
    e.client.add_event_handler(
        #lambda e: test_callback(e,lis),
        cbak,
        events.CallbackQuery(pattern="confirmsetting")
    )

    start = time.time()

    while not lis[0]:
        if (time.time() - start) >= TIMEOUT_SEC:
            break
        await aio.sleep(1)

    val = lis[1]
    
    e.client.remove_event_handler(cbak)

    return val

async def val_input_callback(e,o_sender,lis,file,photo):
    # get the input value
    if o_sender != e.sender_id:
        return
    if not file and not photo:
        lis[0] = True
        lis[1] = e.text
        await e.delete()
    else:
        if e.document is not None and file:
            path = await e.download_media()
            lis[0]  = True
            lis[1] = path 
            await e.delete()
        elif e.photo is not None and photo:
            path = await e.download_media()
            lis[0]  = True
            lis[1] = path 
            await e.delete()
        
        else:
            if "ignore" in e.text:
                lis[0]  = True
                lis[1] = "ignore"
                await e.delete()
            else:
                await e.delete()
        
    raise events.StopPropagation

async def get_confirm_callback(e,o_sender,lis):
    # handle the confirm callback

    if o_sender != e.sender_id:
        return
    lis[0] = True
    
    data = e.data.decode().split(" ")
    if data[1] == "true":
        lis[1] = True
    else:
        lis[1] = False

async def confirm_buttons(e,val):
    # add the confirm buttons at the bottom of the message
    await e.edit(f"confirm the input : <u>{val}</u>",buttons=[KeyboardButtonCallback("yes","confirmsetting true"),KeyboardButtonCallback("no","confirmsetting false")],parse_mode="html")

async def get_bool_variable(var_name,msg,menu,callback_name,sender_id):
    # handle the vars having bool values
     
    val = user_db.get_var(var_name, sender_id)
    try:
        val = bool(val)
    except:
        val = False
    if val:
        #setting the value in callback so calls will be reduced ;)
        menu.append(
            [KeyboardButtonCallback(yes+msg,f"usettings {callback_name} false {sender_id}".encode("UTF-8"))]
        ) 
    else:
        menu.append(
            [KeyboardButtonCallback(no+msg,f"usettings {callback_name} true {sender_id}".encode("UTF-8"))]
        ) 

async def get_sub_menu(msg,sub_name,sender_id,menu):
    menu.append(
        [KeyboardButtonCallback(msg,f"usettings {sub_name} {sender_id}".encode("UTF-8"))]
    )

async def get_string_variable(var_name,menu,callback_name,sender_id):
    # handle the vars having string value
    # condition for rclone config

    if var_name == "RCLONE_CONFIG":
        rfile = user_db.get_rclone(sender_id)
        if rfile is False:
            val = "file didn't load."
        else:
            val = "file loaded"
            #val = "Custom file is loaded."
    else:
        val = user_db.get_var(var_name, sender_id)

       
        
    msg = var_name + " " + str(val)
    menu.append(
        [KeyboardButtonCallback(msg,f"usettings {callback_name} {sender_id}".encode("UTF-8"))]
    )

    # Just in case
    return val

async def get_int_variable(var_name,menu,callback_name,sender_id):
    # handle the vars having string value

    val = user_db.get_var(var_name, sender_id)
    msg = var_name + " " + str(val)
    menu.append(
        [KeyboardButtonCallback(msg,f"usettings {callback_name} {sender_id}".encode("UTF-8"))]
    ) 

# todo handle the list value 


# -*- coding: utf-8 -*-

import asyncio, aria2p, logging, os
from ..core.getVars import get_val
from telethon.tl.types import KeyboardButtonCallback
from telethon.errors.rpcerrorlist import MessageNotModifiedError
from ..core.status.status import ARTask
from functools import partial

# referenced from public leech
# pylint: disable=no-value-for-parameter
torlog = logging.getLogger(__name__)
aloop = asyncio.get_event_loop()

async def aria_start():
    aria2_daemon_start_cmd = []
    # start the daemon, aria2c command
    aria2_daemon_start_cmd.append("aria2c")
    # aria2_daemon_start_cmd.append("--allow-overwrite=true")
    aria2_daemon_start_cmd.append("--daemon=true")
    aria2_daemon_start_cmd.append("--enable-rpc")
    aria2_daemon_start_cmd.append("--disk-cache=0")
    aria2_daemon_start_cmd.append("--follow-torrent=false")
    aria2_daemon_start_cmd.append("--max-connection-per-server=10")
    aria2_daemon_start_cmd.append("--min-split-size=10M")
    aria2_daemon_start_cmd.append("--rpc-listen-all=true")
    aria2_daemon_start_cmd.append(f"--rpc-listen-port=8100")
    aria2_daemon_start_cmd.append("--rpc-max-request-size=1024M")
    aria2_daemon_start_cmd.append("--seed-ratio=0.0")
    aria2_daemon_start_cmd.append("--seed-time=1")
    aria2_daemon_start_cmd.append("--split=10")
    aria2_daemon_start_cmd.append(f"--bt-stop-timeout=100")
    aria2_daemon_start_cmd.append(f"--max-tries=10")
    aria2_daemon_start_cmd.append(f"--retry-wait=2")
    #
    torlog.debug(aria2_daemon_start_cmd)
    #
    process = await asyncio.create_subprocess_exec(
        *aria2_daemon_start_cmd,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE
    )
    stdout, stderr = await process.communicate()
    torlog.debug(stdout)
    torlog.debug(stderr)
    arcli = await aloop.run_in_executor(None, partial(aria2p.Client, host="http://localhost", port=8100, secret=""))
    aria2 = await aloop.run_in_executor(None, aria2p.API, arcli)

    return aria2

async def add_magnet(aria_instance, magnetic_link, c_file_name):
    try:
        download = await aloop.run_in_executor(None, aria_instance.add_magnet, magnetic_link)
    except Exception as e:
        return False, "**failed** \n" + str(e) + " \nplease don't add slow links. read /help."
    else:
        return True, "" + download.gid + ""


async def add_torrent(aria_instance, torrent_file_path):
    if torrent_file_path is None:
        return False, "**failed** \n\nsomething went wrong while adding <u>torrent</u> file."
    if os.path.exists(torrent_file_path):
        # Add Torrent Into Queue
        try:

            download = await aloop.run_in_executor(None, partial(aria_instance.add_torrent, torrent_file_path, uris=None, options=None, position=None))

        except Exception as e:
            return False, "**failed** \n" + str(e) + " \nplease don't add slow links. read /help."
        else:
            return True, "" + download.gid + ""
    else:
        return False, "**failed** \n" + str(e) + " \nplease try other sources to get a working link."


async def add_url(aria_instance, text_url, c_file_name):
    uris = [text_url]
    # Add URL Into Queue
    try:
        
        download = await aloop.run_in_executor(None, aria_instance.add_uris, uris)

    except Exception as e:
        return False, "**failed** \n" + str(e) + " \nplease don't add slow links. read /help."
    else:
        return True, "" + download.gid + ""

async def check_metadata(aria2, gid):
    file = await aloop.run_in_executor(None, aria2.get_download, gid)
    torlog.info(file)
    if not file.followed_by_ids:
        return None
    new_gid = file.followed_by_ids[0]
    torlog.info("changing GID " + gid + " to " + new_gid)
    return new_gid

async def aria_dl(
    incoming_link,
    c_file_name,
    sent_message_to_update_tg_p,
    user_msg
):
    aria_instance = await aria_start()
    
    ar_task = ARTask(None, sent_message_to_update_tg_p, aria_instance, None)
    await ar_task.set_original_mess()

    if incoming_link.lower().startswith("magnet:"):
        sagtus, err_message = await add_magnet(aria_instance, incoming_link, c_file_name)
    elif incoming_link.lower().endswith(".torrent"):
        #sagtus, err_message = await add_torrent(aria_instance, incoming_link)
        #sagtus, err_message = await add_url(aria_instance, incoming_link, c_file_name)
        await ar_task.set_inactive("unable to download .torrent file")
        return False, ar_task
    else:
        sagtus, err_message = await add_url(aria_instance, incoming_link, c_file_name)
    if not sagtus:
        await ar_task.set_inactive(err_message)
        return sagtus, ar_task
        
    torlog.info(err_message)

    await ar_task.set_gid(err_message)

    op = await check_progress_for_dl(
        aria_instance,
        err_message,
        sent_message_to_update_tg_p,
        None,
        ar_task,
        user_msg=user_msg
    )
    if incoming_link.startswith("magnet:"):
        #
        err_message = await check_metadata(aria_instance, err_message)
        #
        await asyncio.sleep(1)
        if err_message is not None:
            op = await check_progress_for_dl(
                aria_instance,
                err_message,
                sent_message_to_update_tg_p,
                None,
                ar_task,
                user_msg=user_msg
            )
        else:
            await ar_task.set_inactive("unable to get metadata.\n")
            return False, ar_task
    await asyncio.sleep(1)
    
    if op is None:
        await ar_task.set_inactive("Known error, nothing went wrong.")
        return False, ar_task
    else:
        statusr, stmsg = op
        if statusr:
            file = await aloop.run_in_executor(None, aria_instance.get_download,err_message)
            to_upload_file = file.name
            await ar_task.set_path(to_upload_file)
            await ar_task.set_done()
            return True, ar_task
        else:
            await ar_task.set_inactive(stmsg)
            return False, ar_task

async def check_progress_for_dl(aria2, gid, event, previous_message, task, rdepth = 0, user_msg=None):
    try:
        file = await aloop.run_in_executor(None, aria2.get_download, gid)
        complete = file.is_complete
        if not complete:
            if not file.error_message:
                msg = ""
                
                await task.refresh_info(file)
                await task.update_message()

                
            else:
                msg = file.error_message
                await event.edit(f"`{msg}`",parse_mode="html", buttons=None)
                torlog.error(f"the aria2 download failed due to: {msg}")
                return False, f"The aria download failed due to: {msg}"
            await asyncio.sleep(get_val("EDIT_SLEEP_SECS"))
            
            # TODO idk not intrested in using recursion here
            return await check_progress_for_dl(
                aria2, gid, event, previous_message,task,user_msg=user_msg
            )
        else:
            await event.edit(f"download completed: `{file.name}` - (`{file.total_length_string()}`)", buttons=None)
            return True, "download complete"
    except aria2p.client.ClientException as e:
        if " not found" in str(e) or "'file'" in str(e):
            fname = "N/A"
            try:
                fname = file.name
            except:pass
            task.cancel = True
            await task.set_inactive()
            return False, f"the download was canceled. {fname}"
        else:
            torlog.warning("errored due to a client error.")
        pass
    except MessageNotModifiedError:
        pass
    except RecursionError:
        file.remove(force=True)
        return False, "The link is dead."
    except Exception as e:
        torlog.info(str(e))
        if " not found" in str(e) or "'file'" in str(e):
            return False, "the download has been already canceled."
        else:
            torlog.warning(str(e))
            return False, f"error: {str(e)}"

async def remove_dl(gid):
    aria2 = await aria_start()
    try:
        downloads = await aloop.run_in_executor(None, aria2.get_download, gid)
        downloads.remove(force=True, files=True)
    except:
        torlog.exception("exc")
        pass



# -*- coding: utf-8 -*-


import psycopg2, logging, time

torlog = logging.getLogger(__name__)

class DataBaseHandle:
    _active_connections = []
    _connection_users = []
    def __init__(self,dburl=None):
        """Load the DB URL if available
        """
        self._dburl = dburl
        if isinstance(self._dburl,bool):
            self._block = True
        else:
            self._block = False
        
        if self._block:
            return
        
        if self._active_connections:
            self._conn = self._active_connections[0]
            self._connection_users.append(1)
        else:
            torlog.debug("Established Connection")
            self._conn = psycopg2.connect(self._dburl)
            self._connection_users.append(1)
            self._active_connections.append(self._conn)

    def scur(self, dictcur=False):
        # start cursor
        cur = None
        for i in range(0,5):
            try:
                if dictcur:
                    cur = self._conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
                else:
                    cur = self._conn.cursor()
                
                break
            except psycopg2.InterfaceError:
                torlog.info(f"Attempting to Re-establish the connection to server {i} times.")
                self.re_establish()

        
        return cur

    def re_establish(self):
        try:
            if self._active_connections[0].closed != 0:
                torlog.info("Re-establish Success.")
                self._conn = psycopg2.connect(self._dburl)
                self._active_connections[0] = self._conn
            else:
                torlog.info("Re-establish Success Cache.")
                self._conn = self._active_connections[0]
        except:
            time.sleep(1) # Blocking call ... this stage is panic.


    def ccur(self,cursor):
        if cursor is not None:
            self._conn.commit()
            cursor.close()


    def __del__(self):
        """Close connection so that the threshold is not exceeded
        """
        if self._block:
            return
        self._connection_users.pop()
        
        if not self._connection_users:
            self._conn.close()


from ... import to_del
from ..getVars import get_val  
import asyncio
import time

async def del_status():
    while True:
        for i in to_del:
            diff = time.time() - i[1]
            if diff >= get_val("STATUS_DEL_TOUT"):
                await i[0].delete()
                to_del.remove(i)

            await asyncio.sleep(1)
        
        await asyncio.sleep(4)


# -*- coding: utf-8 -*-


import logging
import math
import os
import time
from .Human_Format import human_readable_bytes,human_readable_timedelta
from ..core.getVars import get_val
#logging.basicConfig(level=logging.DEBUG)

async def progress(current,total,message,file_name,start,time_out,cancel_msg=None,updb=None):

    now = time.time()
    diff = now - start
    if round(diff % time_out) == 0 or current == total:
        if cancel_msg is not None:
            # dirty alt. was not able to find something to stop upload
            # todo inspect with "StopAsyncIteration"
            if updb.get_cancel_status(cancel_msg.chat_id,cancel_msg.id):
                raise Exception("cancel the upload")

        # if round(current / total * 100, 0) % 5 == 0:
        percentage = current * 100 / total
        speed = current / diff
        elapsed_time = round(diff) * 1000
        time_to_completion = round((total - current) / speed) * 1000
        estimated_total_time = elapsed_time + time_to_completion

        elapsed_time = human_readable_timedelta(seconds=elapsed_time/1000)

        estimated_total_time = human_readable_timedelta(seconds=estimated_total_time/1000)


        progress = "[{0}{1}] \nP: {2}%\n".format(
            ''.join([get_val("COMPLETED_STR") for i in range(math.floor(percentage / 10))]),
            ''.join([get_val("REMAINING_STR") for i in range(10 - math.floor(percentage / 10))]),
            round(percentage, 2))
        
        tmp = progress + "{0} of {1}\nSpeed: {2}/s\nETA: {3}\nusing engine: telethon".format(
            human_readable_bytes(current),
            human_readable_bytes(total),
            human_readable_bytes(speed),
            # elapsed_time if elapsed_time != '' else "0 s",
            estimated_total_time if estimated_total_time != '' else "0 s"
        )
        try:
            if not message.photo:
                await message.edit(
                    text="**Uploading:** `{}`\n{}".format(
                        file_name,
                        tmp
                    )
                )
            else:
                await message.edit(
                    caption="**Uploading:** `{}`\n{}".format(
                        file_name,
                        tmp
                    )
                )
        except Exception as e:
            logging.error(e)
            pass
        return
    else:
        return




class Commands:
    LEECH = "leech"
    PURGE = "purge"
    PAUSEALL = "pauseall"
    RESUMEALL = "resumeall"
    STATUS = "status"
    SETTINGS = "settings"
    EXEC = "exec"
    UPLOAD = "upload"
    YTDL = "ytdl"
    PYTDL = "pytdl"
    ABOUT = "about"
    GETLOGS = "getlogs"
    SERVER = "stats"
    USERSETTINGS = "usettings"
    INSTADL = "instadl"
    START = "start"
    USTATUS = "ustatus"
    SETTHUMB = "setthumb"
    CLRTHUMB = "clearthumb"
    SPEEDTEST = "speedtest"


# -*- coding: utf-8 -*-


from telethon import TelegramClient,events 
from telethon import __version__ as telever
from pyrogram import __version__ as pyrover
from telethon.tl.types import KeyboardButtonCallback
from ..consts.ExecVarsSample import ExecVars
from ..core.getCommand import get_command
from ..core.getVars import get_val
from ..core.speedtest import get_speed
from ..functions.Leech_Module import check_link,cancel_torrent,pause_all,resume_all,purge_all,get_status,print_files, get_transfer
from ..functions.tele_upload import upload_a_file,upload_handel
from ..functions import Human_Format
from .database_handle import tkupload,tktorrents, tkdb
from .settings import handle_settings,handle_setting_callback
from .user_settings import handle_user_settings, handle_user_setting_callback
from functools import partial
from ..functions.rclone_upload import get_config,rclone_driver
from ..functions.admin_check import is_admin
from .. import upload_db, var_db, tor_db, user_db, uptime
import asyncio as aio
import re,logging,time,os,psutil,shutil
from tgtk import __version__
from .ttk_ytdl import handle_ytdl_command,handle_ytdl_callbacks,handle_ytdl_file_download,handle_ytdl_playlist,handle_ytdl_playlist_down
from ..functions.instadl import _insta_post_downloader
torlog = logging.getLogger(__name__)
from .status.status import Status
from .status.menu import create_status_menu, create_status_user_menu
import signal
from PIL import Image

def add_handlers(bot: TelegramClient):
    #bot.add_event_handler(handle_leech_command,events.NewMessage(func=lambda e : command_process(e,get_command("LEECH")),chats=ExecVars.ALD_USR))
    
    bot.add_event_handler(
        handle_leech_command,
        events.NewMessage(pattern=command_process(get_command("LEECH")),
        chats=get_val("ALD_USR"))
    )
    
    bot.add_event_handler(
        handle_purge_command,
        events.NewMessage(pattern=command_process(get_command("PURGE")),
        chats=get_val("ALD_USR"))
    )
    
    bot.add_event_handler(
        handle_pauseall_command,
        events.NewMessage(pattern=command_process(get_command("PAUSEALL")),
        chats=get_val("ALD_USR"))
    )
    
    bot.add_event_handler(
        handle_resumeall_command,
        events.NewMessage(pattern=command_process(get_command("RESUMEALL")),
        chats=get_val("ALD_USR"))
    )

    bot.add_event_handler(
        handle_status_command,
        events.NewMessage(pattern=command_process(get_command("STATUS")),
        chats=get_val("ALD_USR"))
    )

    bot.add_event_handler(
        handle_u_status_command,
        events.NewMessage(pattern=command_process(get_command("USTATUS")),
        chats=get_val("ALD_USR"))
    )

    bot.add_event_handler(
        handle_settings_command,
        events.NewMessage(pattern=command_process(get_command("SETTINGS")),
        chats=get_val("ALD_USR"))
    )

    bot.add_event_handler(
        handle_exec_message_f,
        events.NewMessage(pattern=command_process(get_command("EXEC")),
        chats=get_val("ALD_USR"))
    )
    
    bot.add_event_handler(
        upload_document_f,
        events.NewMessage(pattern=command_process(get_command("UPLOAD")),
        chats=get_val("ALD_USR"))
    )

    bot.add_event_handler(
        handle_ytdl_command,
        events.NewMessage(pattern=command_process(get_command("YTDL")),
        chats=get_val("ALD_USR"))
    )

    bot.add_event_handler(
        handle_ytdl_playlist,
        events.NewMessage(pattern=command_process(get_command("PYTDL")),
        chats=get_val("ALD_USR"))
    )
    
    bot.add_event_handler(
        about_me,
        events.NewMessage(pattern=command_process(get_command("ABOUT")),
        chats=get_val("ALD_USR"))
    )

    bot.add_event_handler(
        get_logs_f,
        events.NewMessage(pattern=command_process(get_command("GETLOGS")),
        chats=get_val("ALD_USR"))
    )
    
    bot.add_event_handler(
        handle_test_command,
        events.NewMessage(pattern="/test",
        chats=get_val("ALD_USR"))
    )

    bot.add_event_handler(
        handle_server_command,
        events.NewMessage(pattern=command_process(get_command("SERVER")),
        chats=get_val("ALD_USR"))
    )
    
    bot.add_event_handler(
        set_password_zip,
        events.NewMessage(pattern=command_process("/setpass"),
        chats=get_val("ALD_USR"))
    )

    bot.add_event_handler(
        handle_user_settings_,
        events.NewMessage(pattern=command_process(get_command("USERSETTINGS")))
    )

    bot.add_event_handler(
        _insta_post_downloader,
        events.NewMessage(pattern=command_process(get_command("INSTADL")),
        chats=get_val("ALD_USR"))
    )

    bot.add_event_handler(
        start_handler,
        events.NewMessage(pattern=command_process(get_command("START")))
    )

    bot.add_event_handler(
        clear_thumb_cmd,
        events.NewMessage(pattern=command_process(get_command("CLRTHUMB")),
        chats=get_val("ALD_USR"))
    )

    bot.add_event_handler(
        set_thumb_cmd,
        events.NewMessage(pattern=command_process(get_command("SETTHUMB")),
        chats=get_val("ALD_USR"))
    )

    bot.add_event_handler(
        speed_handler,
        events.NewMessage(pattern=command_process(get_command("SPEEDTEST")),
        chats=get_val("ALD_USR"))
    )


    signal.signal(signal.SIGINT, partial(term_handler,client=bot))
    signal.signal(signal.SIGTERM, partial(term_handler,client=bot))
    bot.loop.run_until_complete(booted(bot))

    #*********** Callback Handlers *********** 
    
    bot.add_event_handler(
        callback_handler_canc,
        events.CallbackQuery(pattern="torcancel")
    )

    bot.add_event_handler(
        handle_settings_cb,
        events.CallbackQuery(pattern="setting")
    )

    bot.add_event_handler(
        handle_upcancel_cb,
        events.CallbackQuery(pattern="upcancel")
    )

    bot.add_event_handler(
        handle_pincode_cb,
        events.CallbackQuery(pattern="getpin")
    )

    bot.add_event_handler(
        handle_ytdl_callbacks,
        events.CallbackQuery(pattern="ytdlsmenu")
    )

    bot.add_event_handler(
        handle_ytdl_callbacks,
        events.CallbackQuery(pattern="ytdlmmenu")
    )
    
    bot.add_event_handler(
        handle_ytdl_file_download,
        events.CallbackQuery(pattern="ytdldfile")
    )
    
    bot.add_event_handler(
        handle_ytdl_playlist_down,
        events.CallbackQuery(pattern="ytdlplaylist")
    )

    bot.add_event_handler(
        handle_user_setting_callback,
        events.CallbackQuery(pattern="usetting")
    )
    bot.add_event_handler(
        handle_server_command,
        events.CallbackQuery(pattern="fullserver")
    )
#*********** Handlers Below ***********

async def handle_leech_command(e):
    if not e.is_reply:
        await e.reply("reply to a link or magnet.")
    else:
        rclone = False
        tsp = time.time()
        buts = [[KeyboardButtonCallback("to telegram",data=f"leechselect tg {tsp}")]]
        if await get_config() is not None:
            buts.append(
                [KeyboardButtonCallback("to drive",data=f"leechselect drive {tsp}")]
            )
        # tsp is used to split the callbacks so that each download has its own callback
        # cuz at any time there are 10-20 callbacks linked for leeching XD
           
        buts.append(
                [KeyboardButtonCallback("upload in a zip. [Toggle]", data=f"leechzip toggle {tsp}")]
        )
        buts.append(
                [KeyboardButtonCallback("extract from archive. [Toggle]", data=f"leechzipex toggleex {tsp}")]
        )
        
        conf_mes = await e.reply(f"first click if you want to zip the contents or extract as an archive (only one will work at a time) then...\n\n<b>choose where to upload your files:</b>\nthe files will be uploaded to default destination: <b>{get_val('DEFAULT_TIMEOUT')}</b> after 60 sec of no action by user.</u>\n\n<b>supported archives to extract:</b>\nzip, 7z, tar, gzip2, iso, wim, rar, tar.gz, tar.bz2",parse_mode="html",buttons=buts)
        
        # zip check in background
        ziplist = await get_zip_choice(e,tsp)
        zipext = await get_zip_choice(e,tsp,ext=True)
        
        # blocking leech choice 
        choice = await get_leech_choice(e,tsp)
        
        # zip check in backgroud end
        await get_zip_choice(e,tsp,ziplist,start=False)
        await get_zip_choice(e,tsp,zipext,start=False,ext=True)
        is_zip = ziplist[1]
        is_ext = zipext[1]
        
        
        # Set rclone based on choice
        if choice == "drive":
            rclone = True
        else:
            rclone = False
        
        await conf_mes.delete()

        if rclone:
            if get_val("RCLONE_ENABLED"):
                await check_link(e,rclone, is_zip, is_ext)
            else:
                await e.reply("<b>rclone has been disabled by the administrator.</b>",parse_mode="html")
        else:
            if get_val("LEECH_ENABLED"):
                await check_link(e,rclone, is_zip, is_ext)
            else:
                await e.reply("<b>telegram leech has been disabled by the administrator.</b>",parse_mode="html")


async def get_leech_choice(e,timestamp):
    # abstract for getting the confirm in a context

    lis = [False,None]
    cbak = partial(get_leech_choice_callback,o_sender=e.sender_id,lis=lis,ts=timestamp)
    
    gtyh = ""

    e.client.add_event_handler(
        #lambda e: test_callback(e,lis),
        cbak,
        events.CallbackQuery(pattern="leechselect")
    )

    start = time.time()
    defleech = get_val("DEFAULT_TIMEOUT")

    while not lis[0]:
        if (time.time() - start) >= 60: #TIMEOUT_SEC:
            
            if defleech == "leech":
                return "tg"
            elif defleech == "rclone":
                return "drive"
            else:
                # just in case something goes wrong
                return "tg"
            break
        await aio.sleep(1)

    val = lis[1]
    
    e.client.remove_event_handler(cbak)

    return val

async def get_zip_choice(e,timestamp, lis=None,start=True, ext=False):
    # abstract for getting the confirm in a context
    # creating this functions to reduce the clutter
    if lis is None:
        lis = [None, None, None]
    
    if start:
        cbak = partial(get_leech_choice_callback,o_sender=e.sender_id,lis=lis,ts=timestamp)
        lis[2] = cbak
        if ext:
            e.client.add_event_handler(
                cbak,
                events.CallbackQuery(pattern="leechzipex")
            )
        else:
            e.client.add_event_handler(
                cbak,
                events.CallbackQuery(pattern="leechzip")
            )
        return lis
    else:
        e.client.remove_event_handler(lis[2])


async def get_leech_choice_callback(e,o_sender,lis,ts):
    # handle the confirm callback

    if o_sender != e.sender_id:
        return
    data = e.data.decode().split(" ")
    if data [2] != str(ts):
        return
    
    lis[0] = True
    if data[1] == "toggle":
        # encompasses the None situation too
        print("data ",lis)
        if lis[1] is True:
            await e.answer("will not be zipped", alert=True)
            lis[1] = False 
        else:
            await e.answer("will be zipped", alert=True)
            lis[1] = True
    elif data[1] == "toggleex":
        print("exdata ",lis)
        # encompasses the None situation too
        if lis[1] is True:
            await e.answer("it will not be extracted.", alert=True)
            lis[1] = False 
        else:
            await e.answer("if it is a archive, it will be extracted. furthermore, you can add a password if the archive is password protected.", alert=True)
            lis[1] = True
    else:
        lis[1] = data[1]
    

#add admin checks here - done
async def handle_purge_command(e):
    if await is_admin(e.client,e.sender_id,e.chat_id):
        await purge_all(e)
    else:
        await e.delete()

async def handle_pauseall_command(e):
    if await is_admin(e.client,e.sender_id,e.chat_id):
        await pause_all(e)
    else:
        await e.delete()

async def handle_resumeall_command(e):
    if await is_admin(e.client,e.sender_id,e.chat_id):
        await resume_all(e)
    else:
        await e.delete()

async def handle_settings_command(e):
    if await is_admin(e.client,e.sender_id,e.chat_id):
        await handle_settings(e)
    else:
        await e.delete()

async def handle_status_command(e):
    cmds = e.text.split(" ")
    if len(cmds) > 1:
        if cmds[1] == "all":
            await get_status(e,True)
        else:
            await get_status(e)
    else:
        await create_status_menu(e)

async def handle_u_status_command(e):
    await create_status_user_menu(e)
        
async def speed_handler(e):
    if await is_admin(e.client,e.sender_id,e.chat_id):
        await get_speed(e)
        
async def handle_test_command(e):
    pass
    

async def handle_settings_cb(e):
    if await is_admin(e.client,e.sender_id,e.chat_id):
        await handle_setting_callback(e)
    else:
        await e.answer("you do not have the permission to touch this command.",alert=True)

async def handle_upcancel_cb(e):
    db = upload_db

    data = e.data.decode("UTF-8")
    torlog.info("data is {}".format(data))
    data = data.split(" ")

    if str(e.sender_id) == data[3]:
        db.cancel_download(data[1],data[2])
        await e.answer("the upload has been canceled.",alert=True)
    elif e.sender_id in get_val("ALD_USR"):
        db.cancel_download(data[1],data[2])
        await e.answer("UPLOAD CANCELED IN ADMIN MODE XD ;)",alert=True)
    else:
        await e.answer("you can't cancel other peoples uploads.",alert=True)


async def callback_handler_canc(e):
    # TODO the msg can be deleted
    #mes = await e.get_message()
    #mes = await mes.get_reply_message()
    

    torlog.debug(f"the sender id is {e.sender_id}")
    torlog.debug("list bot is authorized on: {} {}".format(get_val("ALD_USR"),type(get_val("ALD_USR"))))

    data = e.data.decode("utf-8").split(" ")
    torlog.debug("data is {}".format(data))
    is_aria = False
    if data[1] == "aria2":
        is_aria = True
        data.remove("aria2")

    if data[2] == str(e.sender_id):
        hashid = data[1]
        hashid = hashid.strip("'")
        torlog.info(f"Hashid :- {hashid}")

        await cancel_torrent(hashid, is_aria)
        await e.answer("the download has been canceled.",alert=True)
    elif e.sender_id in get_val("ALD_USR"):
        hashid = data[1]
        hashid = hashid.strip("'")
        
        torlog.info(f"Hashid: {hashid}")
        
        await cancel_torrent(hashid, is_aria)
        await e.answer("the download has been canceled by an authorized user.",alert=True)
    else:
        await e.answer("you can only cancel your own downloads.", alert=True)


async def handle_exec_message_f(e):
    if get_val("REST11"):
        return
    message = e
    client = e.client
    if await is_admin(client, message.sender_id, message.chat_id, force_owner=True):
        PROCESS_RUN_TIME = 100
        cmd = message.text.split(" ", maxsplit=1)[1]

        reply_to_id = message.id
        if message.is_reply:
            reply_to_id = message.reply_to_msg_id

        process = await aio.create_subprocess_shell(
            cmd,
            stdout=aio.subprocess.PIPE,
            stderr=aio.subprocess.PIPE
        )
        stdout, stderr = await process.communicate()
        e = stderr.decode()
        if not e:
            e = "No Error"
        o = stdout.decode()
        if not o:
            o = "No Output"
        else:
            _o = o.split("\n")
            o = "`\n".join(_o)
        OUTPUT = f"**QUERY:**\n__Command:__\n`{cmd}` \n__PID:__\n`{process.pid}`\n\n**stderr:** \n`{e}`\n**Output:**\n{o}"

        if len(OUTPUT) > 3900:
            with open("exec.text", "w+", encoding="utf8") as out_file:
                out_file.write(str(OUTPUT))
            await client.send_file(
                entity=message.chat_id,
                file="exec.text",
                caption=cmd,
                reply_to=reply_to_id
            )
            os.remove("exec.text")
            await message.delete()
        else:
            await message.reply(OUTPUT)
    else:
        await message.reply("only for owner.")

async def handle_pincode_cb(e):
    data = e.data.decode("UTF-8")
    data = data.split(" ")
    
    if str(e.sender_id) == data[2]:
        db = tor_db
        passw = db.get_password(data[1])
        if isinstance(passw,bool):
            await e.answer("selection is time-sensitive, the download has already started.")
        else:
            await e.answer(f"your pincode is \"{passw}\"",alert=True)

        
    else:
        await e.answer("this torrent is not yours.",alert=True)

async def upload_document_f(message):
    if get_val("REST11"):
        return
    imsegd = await message.reply(
        "processing ..."
    )
    imsegd = await message.client.get_messages(message.chat_id,ids=imsegd.id)
    if await is_admin(message.client, message.sender_id, message.chat_id, force_owner=True):
        if " " in message.text:
            recvd_command, local_file_name = message.text.split(" ", 1)
            recvd_response = await upload_a_file(
                local_file_name,
                imsegd,
                False,
                upload_db
            )
            #torlog.info(recvd_response)
    else:
        await message.reply("only for owner.")
    await imsegd.delete()

async def get_logs_f(e):
    if await is_admin(e.client,e.sender_id,e.chat_id, force_owner=True):
        e.text += " tk_logs.txt"
        await upload_document_f(e)
    else:
        await e.delete()

async def set_password_zip(message):
    #/setpass message_id password
    data = message.raw_text.split(" ")
    passdata = message.client.dl_passwords.get(int(data[1]))
    if passdata is None:
        await message.reply(f"no entry found for this job id {data[1]}")
    else:
        print(message.sender_id)
        print(passdata[0])
        if str(message.sender_id) == passdata[0]:
            message.client.dl_passwords[int(data[1])][1] = data[2]
            await message.reply(f"password updated successfully.")
        else:
            await message.reply(f"cannot update the password since this isn't your download.")

async def start_handler(event):
    msg = "tgtk - a telegram leecher bot."
    await event.reply(msg, parse_mode="html")

def progress_bar(percentage):
    """Returns a progress bar for download
    """
    #percentage is on the scale of 0-1
    comp = get_val("COMPLETED_STR")
    ncomp = get_val("REMAINING_STR")
    pr = ""

    if isinstance(percentage, str):
        return "NaN"

    try:
        percentage=int(percentage)
    except:
        percentage = 0

    for i in range(1,11):
        if i <= int(percentage/10):
            pr += comp
        else:
            pr += ncomp
    return pr

async def handle_server_command(message):
    print(type(message))
    if isinstance(message, events.CallbackQuery.Event):
        callbk = True
    else:
        callbk = False

    try:
        # Memory
        mem = psutil.virtual_memory()
        memavailable = Human_Format.human_readable_bytes(mem.available)
        memtotal = Human_Format.human_readable_bytes(mem.total)
        mempercent = mem.percent
        memfree = Human_Format.human_readable_bytes(mem.free)
    except:
        memavailable = "N/A"
        memtotal = "N/A"
        mempercent = "N/A"
        memfree = "N/A"

    try:
        # Frequencies
        cpufreq = psutil.cpu_freq()
        freqcurrent = cpufreq.current
        freqmax = cpufreq.max
    except:
        freqcurrent = "N/A"
        freqmax = "N/A"

    try:
        # Cores
        cores = psutil.cpu_count(logical=False)
        lcores = psutil.cpu_count()
    except:
        cores = "N/A"
        lcores = "N/A"

    try:
        cpupercent = psutil.cpu_percent()
    except:
        cpupercent = "N/A"
    
    try:
        # Storage
        usage = shutil.disk_usage("/")
        totaldsk = Human_Format.human_readable_bytes(usage.total)
        useddsk = Human_Format.human_readable_bytes(usage.used)
        freedsk = Human_Format.human_readable_bytes(usage.free)
    except:
        totaldsk = "N/A"
        useddsk = "N/A"
        freedsk = "N/A"


    try:
        upb, dlb = await get_transfer()
        dlb = Human_Format.human_readable_bytes(dlb)
        upb = Human_Format.human_readable_bytes(upb)
    except:
        dlb = "N/A"
        upb = "N/A"

    diff = time.time() - uptime
    diff = Human_Format.human_readable_timedelta(diff)

    if callbk:
        msg = (
            f"<b>bot uptime</b> {diff}\n\n"
            "<b>cpu stats:</b>\n"
            f"cores: {cores} logical: {lcores}\n"
            f"cpu frequency: {freqcurrent}  mhz max: {freqmax}\n"
            f"cpu utilization: {cpupercent}%\n"
            "\n"
            "<b>storage stats:</b>\n"
            f"total: {totaldsk}\n"
            f"used: {useddsk}\n"
            f"free: {freedsk}\n"
            "\n"
            "<b>memory stats:</b>\n"
            f"available: {memavailable}\n"
            f"total: {memtotal}\n"
            f"usage: {mempercent}%\n"
            f"free: {memfree}\n"
            "\n"
            "<b>transfer info:</b>\n"
            f"download: {dlb}\n"
            f"upload: {upb}\n"
        )
        await message.edit(msg, parse_mode="html", buttons=None)
    else:
        try:
            storage_percent = round((usage.used/usage.total)*100,2)
        except:
            storage_percent = 0

        
        msg = (
            f"<b>bot uptime:</b> {diff}\n\n"
            f"cpu utilization: {progress_bar(cpupercent)} - {cpupercent}%\n\n"
            f"storage used: {progress_bar(storage_percent)} - {storage_percent}%\n"
            f"total: {totaldsk} free: {freedsk}\n\n"
            f"memory used:- {progress_bar(mempercent)} - {mempercent}%\n"
            f"total: {memtotal} free: {memfree}\n\n"
            f"transfer download: {dlb}\n"
            f"transfer upload: {upb}\n"
        )
        await message.reply(msg, parse_mode="html", buttons=[[KeyboardButtonCallback("get detailed stats.","fullserver")]])


async def about_me(message):
    db = var_db
    _, val1 = db.get_variable("RCLONE_CONFIG")
    if val1 is None:
        rclone_cfg = "no rclone config is loaded."
    else:
        rclone_cfg = "rclone config is loaded."

    val1  = get_val("RCLONE_ENABLED")
    if val1 is not None:
        if val1:
            rclone = "rclone enabled by admin."
        else:
            rclone = "rclone disabled by admin."
    else:
        rclone = "N/A"

    val1  = get_val("LEECH_ENABLED")
    if val1 is not None:
        if val1: 
            leen = "leech command enabled by admin."
        else:
            leen = "leech command disabled by admin."
    else:
        leen = "N/A"


    diff = time.time() - uptime
    diff = Human_Format.human_readable_timedelta(diff)

    msg = (
        "<b>name</b>: <code>tgtk</code>\n"
        f"<b>version</b>: <code>{__version__}</code>\n"
        f"<b>telethon Version</b>: {telever}\n"
        f"<b>pyrogram Version</b>: {pyrover}\n"
        "<u>currents configs:</u>\n\n"
        f"<b>bot uptime:</b> {diff}\n"
        "<b>torrent download engine:</b> <code>qBittorrent [4.3.0 fix active]</code> \n"
        "<b>direct link download engine:</b> <code>aria2</code> \n"
        "<b>upload Engine:</b> <code>rclone</code> \n"
        "<b>youtube download engine:</b> <code>youtube-dl</code>\n"
        f"<b>rclone config: </b> <code>{rclone_cfg}</code>\n"
        f"<b>leech: </b> <code>{leen}</code>\n"
        f"<b>rclone: </b> <code>{rclone}</code>\n"
    )

    await message.reply(msg,parse_mode="html")


async def set_thumb_cmd(e):
    thumb_msg = await e.get_reply_message()
    if thumb_msg is None:
        await e.reply("reply to a photo or photo as a document.")
        return
    
    if thumb_msg.document is not None or thumb_msg.photo is not None:
        value = await thumb_msg.download_media()
    else:
        await e.reply("reply to a photo or photo as a document.")
        return

    try:
        im = Image.open(value)
        im.convert("RGB").save(value,"JPEG")
        im = Image.open(value)
        im.thumbnail((320,320), Image.ANTIALIAS)
        im.save(value,"JPEG")
        with open(value,"rb") as fi:
            data = fi.read()
            user_db.set_thumbnail(data, e.sender_id)
        os.remove(value)
    except Exception:
        torlog.exception("Set Thumb")
        await e.reply("an error occurred while setting the thumbnail.")
        return
    
    try:
        os.remove(value)
    except:pass

    user_db.set_var("DISABLE_THUMBNAIL",False, str(e.sender_id))
    await e.reply("thumbnail set. try using /usettings to get more control. can be used in private too.")

async def clear_thumb_cmd(e):
    user_db.set_var("DISABLE_THUMBNAIL",True, str(e.sender_id))
    await e.reply("thumbnail disabled. try using /usettings to get more control. can be used in private too.")

async def handle_user_settings_(message):
    if not message.sender_id in get_val("ALD_USR"):
        if not get_val("USETTINGS_IN_PRIVATE") and message.is_private:
            return

    await handle_user_settings(message)

def term_handler(signum, frame, client):
    torlog.info("TERM RECEIVD")
    async def term_async():
        omess = None
        st = Status().Tasks
        msg = "the bot is restarting, re-add your tasks.\n\n"
        for i in st:
            if not await i.is_active():
                continue

            omess = await i.get_original_message()
            if str(omess.chat_id).startswith("-100"):
                chat_id = str(omess.chat_id)[4:]
                chat_id = int(chat_id)
            else:
                chat_id = omess.chat_id
            
            sender = await i.get_sender_id()
            msg += f"<a href='tg://user?id={sender}'>reboot</a> - <a href='https://t.me/c/{chat_id}/{omess.id}'>task</a>\n"
        
        if omess is not None:
            await omess.respond(msg, parse_mode="html")
        exit(0)

    client.loop.run_until_complete(term_async())

async def booted(client):
    chats = get_val("ALD_USR")
    for i in chats:
        try:
            await client.send_message(i, "Bot booted successfully!")
        except Exception as e:
            torlog.info(f"Not found the entity {i}")

def command_process(command):
    return re.compile(command,re.IGNORECASE)


# -*- coding: utf-8 -*-


from ..consts.ExecVarsSample import ExecVars
import os
import logging
import time

torlog = logging.getLogger(__name__)

class VarHolder:
    def __init__(self, var_db):
        self._var_dict = dict()
        self._vardb = var_db

    def get_var(self, variable):
        if variable in self._var_dict.keys():
            torlog.debug("network call no made")
            return self._var_dict[variable]
        torlog.debug("network call made")
        db = self._vardb
        val = None

        #Get the variable from the constants supplied
        try:
            val = getattr(ExecVars,variable)
        except AttributeError:pass

        #Get the variable form the env [overlap]
        #try:
        envval = os.environ.get(variable)
        INTS = ["EDIT_SLEEP_SECS", "MAX_TORRENT_SIZE", "MAX_YTPLAYLIST_SIZE", "TG_UP_LIMIT", "API_ID", "STATUS_DEL_TOUT", "TOR_MAX_TOUT", "OWNER_ID"]
        BOOLS = ["FORCE_DOCUMENTS", "LEECH_ENABLED", "RCLONE_ENABLED", "USETTINGS_IN_PRIVATE"]

        if variable == "ALD_USR":
            if envval is not None:
                templi = envval.split(" ")
                templi2 = []
                if len(templi) > 0:
                    for i in range(0,len(templi)):
                        try:
                            templi2.append(int(templi[i]))
                        except ValueError:
                            torlog.error(f"Invalid allow user {templi[i]} must be a integer.")
                if val is not None:
                    val.extend(templi2)
                else:
                    val = templi
        elif variable in INTS:
            val =  int(envval) if envval is not None else val
        elif variable in BOOLS:
            if envval:
                if not isinstance(val, bool):
                    if "true" in envval.lower():
                        val = True
                    else:
                        val = False
            else:
                val = None
        else:
            val =  envval if envval is not None else val

        #Get the variable form the DB [overlap]
        dbval, _ = db.get_variable(variable)
        
        if dbval is not None:
            val = dbval

        if val is None:
            torlog.error("The variable was not found in either the constants, environment or database. Variable is :- {}".format(variable))
            #raise Exception("The variable was not found in either the constants, environment or database. Variable is :- {}".format(variable))
        
        if isinstance(val,str):
            val = val.strip()

        self._var_dict[variable] = val
        return val

    def update_var(self, name, val):
        self._var_dict[name] = val 


# -*- coding: utf-8 -*-


from anytree import NodeMixin, RenderTree, PreOrderIter
import qbittorrentapi as qba
from ..functions.Human_Format import human_readable_bytes
#from ..functions.Human_Format import human_readable_bytes

class TorNode(NodeMixin):
    def __init__(self,name,is_folder=False,is_file=False,parent=None,progress=None,size=None,priority=None,file_id=None):
        super().__init__()
        self.name = name
        self.is_folder = is_folder
        self.is_file = is_file
        
        if parent is not None:
            self.parent = parent
        if progress is not None:
            self.progress = progress
        if size is not None:
            #self.size = human_readable_bytes(size)
            self.size = size
        if priority is not None:
            self.priority = priority
        if file_id is not None:
            self.file_id = file_id
        

def get_folders(path):
    path_seperator = "/"
    folders = path.split(path_seperator)
    return folders


def make_tree(res):
    """This function takes the list of all the torrent files. The files are name hierarchically.
        Felt a need to document to save time.

    Args:
        res (list): Torrent files list.

    Returns:
        TorNode: Parent node of the tree constructed and can be used further.
    """
    parent = TorNode("Torrent")
    #nodes = dict()
    l = 0
    
    for i in res:
        # Get the hierarchy of the folders by splitting based on '/'
        folders = get_folders(i.name)
        # Check if the file is alone for if its in folder
        if len(folders) > 1:
            # Enter here if in folder

            # Set the parent 
            previous_node = parent
            
            # Traverse till second last assuming the last is a file.
            for j in range(len(folders)-1):
                current_node = None
                
                if previous_node is not None:
                    # As we are traversing the folder from top to bottom we are searching
                    # the first folder (folders list) under the parent node in first iteration.
                    # If the node is found then it becomes the current node else the current node
                    # is left None. 
                    for k in previous_node.children:
                        if k.name == folders[j]:
                            current_node = k 
                            break
                else:
                    # think its useless afterall
                    for k in parent.children:
                        if k.name == folders[j]:
                            current_node = k
                            break
                
                # if the node is not found then create the folder node
                # if the node is found then use it as base for the next 
                if current_node is None:
                    previous_node = TorNode(folders[j],parent=previous_node,is_folder=True)
                else:
                    previous_node = current_node
            # at this point the previous_node will contain the deepest folder in it so add the file to it 
            TorNode(folders[-1],is_file=True,parent=previous_node,progress=i.progress,size=i.size,priority=i.priority,file_id=l)
            l += 1
        else:
            # at the file to the parent if no folders are there 
            TorNode(folders[-1],is_file=True,parent=parent,progress=i.progress,size=i.size,priority=i.priority,file_id=l)
            l += 1


    return parent

    # Will be depricated after testing 
    #for i in res:
    #    folders = get_folders(i.name)
    #    if len(folders) > 1:
    #        for j in range(len(folders)-1):
    #            if not (folders[j] in nodes.keys()):
    #                if j != 0:
    #                    nodes[folders[j]] = TorNode(folders[j],True,parent=nodes[folders[j-1]])
    #                else:
    #                    nodes[folders[j]] = TorNode(folders[j],True,parent=parent)
    #        TorNode(folders[-1],is_file=True,parent=nodes[folders[-2]],progress=i.progress,size=i.size,priority=i.priority,file_id=k)
    #        k += 1
    #    else:
    #        TorNode(folders[-1],is_file=True,parent=parent,progress=i.progress,size=i.size,priority=i.priority,file_id=k)
    #        k += 1
    #return parent

def print_tree(parent):
    for pre, _, node in RenderTree(parent):
        treestr = u"%s%s" % (pre, node.name)
        print(treestr.ljust(8), node.is_folder, node.is_file)


def create_list(par,msg):
    if par.name != ".unwanted":
        msg[0] += "<ul>"
    for i in par.children:
        if i.is_folder:
            msg[0] += "<li>"
            if i.name != ".unwanted":
                msg[0] += f"<input type=\"checkbox\" name=\"foldernode_{msg[1]}\"> <label for=\"{i.name}\">{i.name}</label>"
            create_list(i,msg)
            msg[0] += "</li>"
            msg[1] += 1
        else:
            msg[0] += "<li>"
            if i.priority == 0:
                msg[0] += f"<input type=\"checkbox\" name=\"filenode_{i.file_id}\"> <label for=\"filenode_{i.file_id}\">{i.name} - {human_readable_bytes(i.size)}</label>"
                msg[0] += f"<input type=\"hidden\" value=\"off\" name=\"filenode_{i.file_id}\">"
                
            else:
                msg[0] += f"<input type=\"checkbox\" checked name=\"filenode_{i.file_id}\"> <label for=\"filenode_{i.file_id}\">{i.name} - {human_readable_bytes(i.size)}</label>"
                msg[0] += f"<input type=\"hidden\" value=\"off\" name=\"filenode_{i.file_id}\">"
            

            msg[0] += "</li>"
    
    if par.name != ".unwanted":
        msg[0] += "</ul>"




import re
import aiohttp
import math, json
import urllib.parse
import logging
from bs4 import BeautifulSoup

torlog = logging.getLogger(__name__)

async def generate_directs(url):
    #blocklinks
    if 'drive.google.com' in url or 'uptobox.com' in url \
    or '1fiecher.com' in url or 'googleusercontent.com' in url:
        return "**ERROR:** Unsupported URL!"
    
    #mediafire.com
    elif 'mediafire.com' in url:
        try:
            link = re.findall(r'\bhttps?://.*mediafire\.com\S+', url)[0]
            async with aiohttp.ClientSession() as ttksess:
                resp = await ttksess.get(link)
                restext = await resp.text()

            page = BeautifulSoup(restext, 'lxml')
            info = page.find('a', {'aria-label': 'Download file'})
            ourl = info.get('href')
            return ourl
        except:
            return "**ERROR:** can't download, double check your mediafire link!"
          
    #disk.yandex.com    
    elif 'yadi.sk' in url or 'disk.yandex.com' in url:
        try:
            link = re.findall(r'\b(https?://.*(yadi|disk)\.(sk|yandex)*(|com)\S+)', url)[0][0]
            print(link)
        except:
            return "**ERROR:** Cant't download, double check your yadisk link!"

        api = 'https://cloud-api.yandex.net/v1/disk/public/resources/download?public_key={}'
        try:
            async with aiohttp.ClientSession() as ttksess:
                resp = await ttksess.get(api.format(link))
                restext = await resp.json()
                ourl = restext['href']
                return ourl
        except:
            torlog.exception("Ayee jooo")
            return "**ERROR:** can't download, the yadisk file not found or dowmload limit reached!" 

    #zippyshare.com
    elif 'zippyshare.com' in url:
        try:
            link = re.findall(r'\bhttps?://.*zippyshare\.com\S+', url)[0]
            async with aiohttp.ClientSession() as ttksess:
                resp = await ttksess.get(link)
                restext = await resp.text()
            base_url = re.search('http.+.com', restext).group()
            page_soup = BeautifulSoup(restext, 'lxml')
            scripts = page_soup.find_all("script", {"type": "text/javascript"})
            for script in scripts:
                if "getElementById('dlbutton')" in script.text:
                    url_raw = re.search(r'= (?P<url>\".+\" \+ (?P<math>\(.+\)) .+);',
                                        script.text).group('url')
                    math = re.search(r'= (?P<url>\".+\" \+ (?P<math>\(.+\)) .+);',
                       script.text).group('math')
                    url = url_raw.replace(math, '"' + str(eval(math)) + '"')
                    break
            ourl = base_url + eval(url)
            name = urllib.parse.unquote(url.split('/')[-1])
            return ourl
        except:
            return "**ERROR:** can't download, double check your zippyshare link!"
       
    #racaty.net
    elif 'racaty.net' in url:
        try:
            link = re.findall(r'\bhttps?://.*racaty\.net\S+', url)[0]
            async with aiohttp.ClientSession() as ttksess:
                resp = await ttksess.get(link)
                restext = await resp.text()                      
            bss=BeautifulSoup(restext,'html.parser')
            op=bss.find('input',{'name':'op'})['value']
            id=bss.find('input',{'name':'id'})['value']
                  
            async with aiohttp.ClientSession() as ttksess:
                rep = await ttksess.post(link,data={'op':op,'id':id})
                reptext = await rep.text()
            bss2=BeautifulSoup(reptext,'html.parser')
            ourl = bss2.find('a',{'id':'uniqueExpirylink'})['href']
            return ourl
        except:
            return "**ERROR:** can't download, double check your racaty link!"
       
    elif 'pixeldrain.com' in url:
        url = url.strip("/ ")
        file_id = url.split("/")[-1]
        
        info_link = f"https://pixeldrain.com/api/file/{file_id}/info"
        dl_link = f"https://pixeldrain.com/api/file/{file_id}"

        async with aiohttp.ClientSession() as ttksess:
            resp = await ttksess.get(info_link)
            restext = await resp.json()

        if restext["success"]:
            return dl_link
        else:
            return "**ERROR:** can't download, {}.".format(restext["value"])


# -*- coding: utf-8 -*-


import qbittorrentapi as qba
import asyncio as aio
import os,logging,traceback,time
from datetime import datetime,timedelta
from . import Hash_Fetch
from .Human_Format import human_readable_bytes,human_readable_timedelta 
from ..core.getVars import get_val
from telethon.tl.types import KeyboardButtonCallback,KeyboardButtonUrl
from telethon.errors.rpcerrorlist import MessageNotModifiedError,FloodWaitError
from telethon import events
from functools import partial, wraps
from random import randint
from .. import tor_db
from ..core.status.status import QBTask

#logging.basicConfig(level=logging.DEBUG)
torlog = logging.getLogger(__name__)
aloop = aio.get_event_loop()
logging.getLogger('qbittorrentapi').setLevel(logging.ERROR)
logging.getLogger('requests').setLevel(logging.ERROR)
logging.getLogger('urllib3').setLevel(logging.ERROR)

async def get_client(host=None,port=None,uname=None,passw=None,retry=2) -> qba.TorrentsAPIMixIn:
    """Creats and returns a client to communicate with qBittorrent server. Max Retries 2
    """
    #getting the conn 
    host = host if host is not None else "localhost"
    port = port if port is not None else "8090"
    uname = uname if uname is not None else "admin"
    passw = passw if passw is not None else "adminadmin"
    torlog.info(f"Trying to login in qBittorrent using creds {host} {port} {uname} {passw}")

    client = qba.Client(host=host,port=port,username=uname,password=passw)
    
    #try to connect to the server :)
    try:
        await aloop.run_in_executor(None,client.auth_log_in)
        torlog.info("Client connected successfully to the torrent server. :)")
        await aloop.run_in_executor(None,client.application.set_preferences,{"disk_cache":20,"incomplete_files_ext":True,"max_connec":3000,"max_connec_per_torrent":300,"async_io_threads":6})
        torlog.debug("Setting the cache size to 20 incomplete_files_ext:True,max_connec:3000,max_connec_per_torrent:300,async_io_threads:6")
        return client
    except qba.LoginFailed as e:
        torlog.error("An errot occured invalid creds detected\n{}\n{}".format(e,traceback.format_exc()))
        return None
    except qba.APIConnectionError:
        if retry == 0:
            torlog.error("Tried to get the client 3 times no luck")
            return None
        
        torlog.info("Oddly enough the qbittorrent server is not running.... Attempting to start at port {}".format(port))
        cmd = f"qbittorrent-nox -d --webui-port={port}"
        cmd = cmd.split(" ")

        subpr = await aio.create_subprocess_exec(*cmd,stderr=aio.subprocess.PIPE,stdout=aio.subprocess.PIPE)
        _, _ = await subpr.communicate()
        return await get_client(host,port,uname,passw,retry=retry-1)


async def add_torrent_magnet(magnet,message):
    """Adds a torrent by its magnet link.
    """    
    client = await get_client()
    try:
        ctor = len(await get_torrent_info(client))
        
        ext_hash = Hash_Fetch.get_hash_magnet(magnet)
        ext_res = await get_torrent_info(client, ext_hash)

        if len(ext_res) > 0:
            torlog.info(f"this torrent is in list {ext_res} {magnet} {ext_hash}")
            await message.edit("this torrent is already added to the leech queue.")
            return False
        # hot fix for the below issue
        savepath = os.path.join(os.getcwd(), "Downloads", str(time.time()).replace(".",""))
        op = await aloop.run_in_executor(None,partial(client.torrents_add, magnet, save_path=savepath))
        
        
        # TODO uncomment the below line and remove the above fix when fixed https://github.com/qbittorrent/qBittorrent/issues/13572
        # op = client.torrents_add(magnet)

        # torrents_add method dosent return anything so have to work around
        if op.lower() == "ok.":
            st = datetime.now()
            
            ext_res = await get_torrent_info(client, ext_hash)
            if len(ext_res) > 0:
                torlog.info("Got torrent info from ext hash.")
                return ext_res[0]

            while True:
                if (datetime.now() - st).seconds >= 10:
                    torlog.warning("The provided torrent was not added and it was timed out. magnet was:- {}".format(magnet))
                    torlog.error(ext_hash)
                    await message.edit("The torrent was not added due to an error.")
                    return False
                # commenting in favour of wrong torrent getting returned
                # ctor_new = client.torrents_info()
                #if len(ctor_new) > ctor:
                #    # https://t.me/c/1439207386/2977 below line is for this
                #    torlog.info(ctor_new)
                #    torlog.info(magnet)
                #    return ctor_new[0]
                ext_res = await get_torrent_info(client, ext_hash)
                if len(ext_res) > 0:
                    torlog.info("Got torrent info from ext hash.")
                    return ext_res[0]

        else:
            await message.edit("This is an unsupported/invalid link.")
    except qba.UnsupportedMediaType415Error as e:
        #will not be used ever ;)
        torlog.error("Unsupported file was detected in the magnet here")
        await message.edit("This is an unsupported/invalid link.")
        return False
    except Exception as e:
        torlog.error("{}\n{}".format(e,traceback.format_exc()))
        await message.edit("Error occured check logs.")
        return False

async def add_torrent_file(path,message):
    if not os.path.exists(path):
        torlog.error("The path supplied to the torrent file was invalid.\n path:-{}".format(path))
        return False

    client = await get_client()
    try:
        ctor = len(await get_torrent_info(client))

        ext_hash = Hash_Fetch.get_hash_file(path)
        ext_res = await get_torrent_info(client, ext_hash)

        if len(ext_res) > 0:
            torlog.info(f"this torrent is in list {ext_res} {path} {ext_hash}")
            await message.edit("this torrent is already added to the leech queue")
            return False
        
        # hot fix for the below issue
        savepath = os.path.join(os.getcwd(), "Downloads", str(time.time()).replace(".",""))

        op = await aloop.run_in_executor(None,partial(client.torrents_add,torrent_files=[path], save_path=savepath))
        
        # TODO uncomment the below line and remove the above fix when fixed https://github.com/qbittorrent/qBittorrent/issues/13572
        # op = client.torrents_add(torrent_files=[path])
        #this method dosent return anything so have to work around
        
        if op.lower() == "ok.":
            st = datetime.now()
            #ayehi wait karna hai
            await aio.sleep(2)
            
            ext_res = await get_torrent_info(client, ext_hash)
            if len(ext_res) > 0:
                torlog.info("Got torrent info from ext hash.")
                return ext_res[0]

            while True:
                if (datetime.now() - st).seconds >= 20:
                    torlog.warning("The provided torrent was not added and it was timed out. file path was:- {}".format(path))
                    torlog.error(ext_hash)
                    await message.edit("The torrent was not added due to an error.")
                    return False
                #ctor_new = client.torrents_info()
                #if len(ctor_new) > ctor:
                #    return ctor_new[0]
                ext_res = await get_torrent_info(client, ext_hash)
                if len(ext_res) > 0:
                    torlog.info("Got torrent info from ext hash.")
                    return ext_res[0]

        else:
            await message.edit("This is an unsupported/invalid link.")
    except qba.UnsupportedMediaType415Error as e:
        #will not be used ever ;)
        torlog.error("Unsupported file was detected in the magnet here")
        await message.edit("This is an unsupported/invalid link.")
        return False
    except Exception as e:
        torlog.error("{}\n{}".format(e,traceback.format_exc()))
        await message.edit("Error occured check logs.")
        return False

async def update_progress(client,message,torrent,task,except_retry=0,sleepsec=None):
    #task = QBTask(torrent, message, client)
    if sleepsec is None:
        sleepsec = get_val("EDIT_SLEEP_SECS")
    #switch to iteration from recursion as python dosent have tailing optimization :O
    #RecursionError: maximum recursion depth exceeded
    is_meta = False
    meta_time = time.time()

    while True:
        tor_info = await get_torrent_info(client, torrent.hash)
        #update cancellation
        if len(tor_info) > 0:
            tor_info = tor_info[0]
        else:
            task.cancel = True
            await task.set_inactive()
            await message.edit("Torrent canceled ```{}``` ".format(torrent.name),buttons=None)
            return True
        
        if tor_info.size > (get_val("MAX_TORRENT_SIZE") * 1024 * 1024 * 1024):
            await message.edit("Torrent oversized max size is {}. Try adding again and choose less files to download.".format(get_val("MAX_TORRENT_SIZE")), buttons=None)
            await delete_this(tor_info.hash)
            return True
        try:
            await task.refresh_info(tor_info)
            await task.update_message()

            if  tor_info.state == "metaDL":
                is_meta = True
            else:
                meta_time = time.time()
                is_meta = False

            if (is_meta and (time.time() - meta_time) > get_val("TOR_MAX_TOUT")):
                
                await message.edit("Torrent <code>{}</code> is DEAD. [Metadata Failed]".format(tor_info.name),buttons=None,parse_mode="html")
                torlog.error("An torrent has no seeds clearing that torrent now. Torrent:- {} - {}".format(tor_info.hash,tor_info.name))
                await delete_this(tor_info.hash)
                await task.set_inactive("Torrent <code>{}</code> is DEAD. [Metadata Failed]".format(tor_info.name))
                
                return False

            try:
                if tor_info.state == "error":

                    await message.edit("Torrent <code>{}</code> errored out.".format(tor_info.name),buttons=None,parse_mode="html")
                    torlog.error("An torrent has error clearing that torrent now. Torrent:- {} - {}".format(tor_info.hash,tor_info.name))
                    await delete_this(tor_info.hash)
                    await task.set_inactive("Torrent <code>{}</code> errored out.".format(tor_info.name))
                    
                    return False
                
                #aio timeout have to switch to global something
                await aio.sleep(sleepsec)

                #stop the download when download complete
                if tor_info.state == "uploading" or tor_info.state.lower().endswith("up"):
                    # this is to address the situations where the name would cahnge abdruptly
                    await aloop.run_in_executor(None,partial(client.torrents_pause,tor_info.hash))

                    # TODO uncomment the below line when fixed https://github.com/qbittorrent/qBittorrent/issues/13572
                    # savepath = os.path.join(tor_info.save_path,tor_info.name)
                    # hot fix
                    try:
                        savepath = os.path.join(tor_info.save_path, os.listdir(tor_info.save_path)[-1])
                    except:
                        await message.edit("Download path location failed", buttons=None)
                        await task.set_inactive("Download path location failed")
                        await delete_this(tor_info.hash)
                        return None

                    await task.set_path(savepath)
                    await task.set_done()
                    await message.edit("download completed: `{}` - (`{}`)\nto path: `{}`".format(tor_info.name,human_readable_bytes(tor_info.total_size),tor_info.save_path),buttons=None)
                    return [savepath, task]
                else:
                    #return await update_progress(client,message,torrent)
                    pass

            except (MessageNotModifiedError,FloodWaitError) as e:
                torlog.error("{}".format(e))
            
        except Exception as e:
            torlog.error("{}\n\n{}\n\nn{}".format(e,traceback.format_exc(),tor_info))
            try:
                await message.edit("Error occure {}".format(e),buttons=None)
            except:pass
            return False

async def pause_all(message):
    client = await get_client()
    await aloop.run_in_executor(None,partial(client.torrents_pause,torrent_hashes='all'))
    await aio.sleep(1)
    msg = ""
    tors = await aloop.run_in_executor(None,partial(client.torrents_info,status_filter="paused|stalled"))
    msg += "⏸️ Paused total <b>{}</b> torrents ⏸️\n".format(len(tors))

    for i in tors:
        if i.progress == 1:
            continue
        msg += "➡️<code>{}</code> - <b>{}%</b>\n".format(i.name,round(i.progress*100,2))

    await message.reply(msg,parse_mode="html")
    await message.delete()

async def resume_all(message):
    client = await get_client()

    await aloop.run_in_executor(None,partial(client.torrents_resume, torrent_hashes='all'))

    await aio.sleep(1)
    msg = ""
    tors = await aloop.run_in_executor(None,partial(client.torrents_info,status_filter="stalled|downloading|stalled_downloading"))
    
    msg += "▶️Resumed {} torrents check the status for more...▶️".format(len(tors))

    for i in tors:
        if i.progress == 1:
            continue
        msg += "➡️<code>{}</code> - <b>{}%</b>\n".format(i.name,round(i.progress*100,2))

    await message.reply(msg,parse_mode="html")
    await message.delete()

async def delete_all(message):
    client = await get_client()
    tors = await get_torrent_info(client)
    msg = "☠️ Deleted <b>{}</b> torrents.☠️".format(len(tors))
    client.torrents_delete(delete_files=True,torrent_hashes="all")

    await message.reply(msg,parse_mode="html")
    await message.delete()
    
async def delete_this(ext_hash):
    client = await get_client()
    await aloop.run_in_executor(None,partial(client.torrents_delete,delete_files=True,torrent_hashes=ext_hash))
    return True

async def get_status(message,all=False):
    client = await get_client()
    tors = await get_torrent_info(client)
    olen = 0

    if len(tors) > 0:
        msg = ""
        for i in tors:
            if i.progress == 1 and not all:
                continue
            else:
                olen += 1
                msg += "📥 <b>{} | {}% | {}/{}({}) | {} | {} | S:{} | L:{} | {}</b>\n\n".format(
                    i.name,
                    round(i.progress*100,2),
                    human_readable_bytes(i.completed),
                    human_readable_bytes(i.size),
                    human_readable_bytes(i.total_size),
                    human_readable_bytes(i.dlspeed,postfix="/s"),
                    human_readable_timedelta(i.eta),
                    i.num_seeds,
                    i.num_leechs,
                    i.state
                )
        if msg.strip() == "":
            return "No torrents running currently...." 
        return msg
    else:
        msg = "No torrents running currently...."
        return msg
    
    if olen == 0:
        msg = "No torrents running currently...."
        return msg



def progress_bar(percentage):
    """Returns a progress bar for download
    """
    #percentage is on the scale of 0-1
    comp = get_val("COMPLETED_STR")
    ncomp = get_val("REMAINING_STR")
    pr = ""

    for i in range(1,11):
        if i <= int(percentage*10):
            pr += comp
        else:
            pr += ncomp
    return pr

async def deregister_torrent(hashid):
    client = await get_client()
    await aloop.run_in_executor(None,partial(client.torrents_delete, torrent_hashes=hashid,delete_files=True))

async def register_torrent(entity,message,user_msg=None,magnet=False,file=False):
    client = await get_client()

    #refresh message
    message = await message.client.get_messages(message.chat_id,ids=message.id)
    if user_msg is None:
        omess = await message.get_reply_message()
    else:
        omess = user_msg

    if magnet:
        torlog.info(f"magnet :- {magnet}")
        torrent = await add_torrent_magnet(entity,message)
        if isinstance(torrent,bool):
            return False
        torlog.info(torrent)
        if torrent.progress == 1 and torrent.completion_on > 1:
            await message.edit("The provided torrent was already completly downloaded.")
            return True
        else:
            
            pincode = randint(1000,9999)
            db = tor_db
            db.add_torrent(torrent.hash,pincode)
            
            pincodetxt = f"getpin {torrent.hash} {omess.sender_id}"

            data = "torcancel {} {}".format(torrent.hash, omess.sender_id)
            base = get_val("BASE_URL_OF_BOT")

            urll = f"{base}/tgtk/files/{torrent.hash}"

            message = await message.edit("Download will be automatically started after 180s of no action.",buttons=[
                [
                    KeyboardButtonUrl("Choose File from link",urll),
                    KeyboardButtonCallback("Get Pincode",data=pincodetxt.encode("UTF-8"))
                ],
                [
                    KeyboardButtonCallback("Done Selecting Files.",data=f"doneselection {omess.sender_id} {omess.id}".encode("UTF-8"))
                ]
            ])

            await get_confirm(omess)

            message = await message.edit(buttons=[KeyboardButtonCallback("Cancel Leech",data=data.encode("UTF-8"))])

            db.disable_torrent(torrent.hash)
            

            task = QBTask(torrent, message, client)
            await task.set_original_mess(omess)
            return await update_progress(client,message,torrent, task)
    if file:
        torrent = await add_torrent_file(entity,message)
        if isinstance(torrent,bool):
            return False
        torlog.info(torrent)
        
        if torrent.progress == 1:
            await message.edit("The provided torrent was already completly downloaded.")
            return True
        else:
            pincode = randint(1000,9999)
            db = tor_db
            db.add_torrent(torrent.hash,pincode)
            
            pincodetxt = f"getpin {torrent.hash} {omess.sender_id}"

            data = "torcancel {} {}".format(torrent.hash, omess.sender_id)

            base = get_val("BASE_URL_OF_BOT")

            urll = f"{base}/tgtk/files/{torrent.hash}"

            message = await message.edit(buttons=[
                [
                    KeyboardButtonUrl("Choose File from link",urll),
                    KeyboardButtonCallback("Get Pincode",data=pincodetxt.encode("UTF-8"))
                ],
                [
                    KeyboardButtonCallback("Done Selecting Files.",data=f"doneselection {omess.sender_id} {omess.id}".encode("UTF-8"))
                ]
            ])

            await get_confirm(omess)

            message = await message.edit(buttons=[KeyboardButtonCallback("Cancel Leech",data=data.encode("UTF-8"))])

            db.disable_torrent(torrent.hash)
            
            task = QBTask(torrent, message, client)
            await task.set_original_mess(omess)
            return await update_progress(client,message,torrent, task)

async def get_confirm(e):
    # abstract for getting the confirm in a context

    lis = [False,None,e.id]
    cbak = partial(get_confirm_callback,lis=lis)
    
    e.client.add_event_handler(
        #lambda e: test_callback(e,lis),
        cbak,
        events.CallbackQuery(pattern="doneselection")
    )

    start = time.time()

    while not lis[0]:
        if (time.time() - start) >= 180:
            break
        await aio.sleep(1)

    val = lis[1]
    
    e.client.remove_event_handler(cbak)

    return val

async def get_confirm_callback(e,lis):
    # handle the confirm callback
    data = e.data.decode("UTF-8")
    data = data.split(" ")
    o_sender = data[1]
    msgid = data[2]

    if o_sender != str(e.sender_id):
        await e.answer("Dont Touch it.......")
        return
    if str(lis[2]) != msgid:
        return
    await e.answer("Starting the download with the selected files.")
    lis[0] = True
    raise events.StopPropagation()

# quick asycn functions

async def get_torrent_info(client, ehash=None):

    if ehash is None:
        return await aloop.run_in_executor(None,client.torrents_info)
    else:
        return await aloop.run_in_executor(None,partial(client.torrents_info,torrent_hashes=ehash))




# -*- coding: utf-8 -*-

from .status import Status, QBTask, ARTask
from .upload import TGUploadTask, RCUploadTask
from telethon.tl.types import KeyboardButtonCallback
from ... import to_del
import time, asyncio, logging

torlog = logging.getLogger(__name__)

def get_num(no):
    nums = ['0️⃣','1️⃣','2️⃣','3️⃣','4️⃣','5️⃣','6️⃣','7️⃣','8️⃣','9️⃣','🔟']
    numstr = ""

    if no <= 9:
        return nums[no]
    else:
        for i in str(no):
            numstr += nums[int(i)]

    return numstr

async def create_status_menu(event):
    
    tasks = Status()
    tors = 0
    row = []
    Buttons = []

    msg = "currently running: \nclick on the task number you want to cancel.\n"
    for i in tasks.Tasks:
        if await i.is_active():
            
            msg += get_num(tors) + " " + await i.create_message()
            msg += "\n\n"
            try:
                if isinstance(i, QBTask):
                    omsg = await i.get_original_message()
                    data = "torcancel {} {}".format(
                        i.hash, 
                        omsg.sender_id
                    )
                if isinstance(i, ARTask):
                    data = "torcancel aria2 {} {}".format(
                        await i.get_gid(),
                        await i.get_sender_id()
                    )
                if isinstance(i, TGUploadTask):
                    message = await i.get_message()
                    data = "upcancel {} {} {}".format(
                        message.chat_id,
                        message.id,
                        await i.get_sender_id()
                    )
                if isinstance(i, RCUploadTask):
                    omsg = await i.get_original_message()
                    data = "upcancel {} {} {}".format(
                        omsg.chat_id,
                        omsg.id,
                        omsg.sender_id
                    )
            except:
                torlog.exception("In status msg")
                tors += 1
                continue
            
            row.append(KeyboardButtonCallback(get_num(tors), data=data.encode("UTF-8")))
            if len(row) >= 4:
                Buttons.append(row)
                row = []
            
            tors += 1

    if row:
        Buttons.append(row)

    if not Buttons:
        Buttons = None
    
    if len(msg) > 3900:
        chunks, chunk_size = len(msg), 3900
        msgs = [ msg[i:i+chunk_size] for i in range(0, chunks, chunk_size) ]
        
        for j in msgs:
            memsg = await event.reply(j,parse_mode="html",  buttons=Buttons)
            to_del.append([memsg, time.time()])
            await asyncio.sleep(1)
    else:
        memsg = await event.reply(msg,parse_mode="html",  buttons=Buttons)
        to_del.append([memsg, time.time()])
    #memsg = await event.reply(msg,parse_mode="html", buttons=Buttons)
    #to_del.append([memsg, time.time()])

async def create_status_user_menu(event):
    
    tasks = Status()
    tors = 0
    row = []
    Buttons = []

    msg = "currently running: for you- \nclick on the task number you want to cancel.\n"
    for i in tasks.Tasks:
        if await i.is_active():
            
            try:
                if isinstance(i, QBTask):
                    omsg = await i.get_original_message()
                    if not (event.sender_id == omsg.sender_id):
                        continue
                    data = "torcancel {} {}".format(
                        i.hash, 
                        omsg.sender_id
                    )
                if isinstance(i, ARTask):
                    if not (event.sender_id == await i.get_sender_id()):
                        continue
                    data = "torcancel aria2 {} {}".format(
                        await i.get_gid(),
                        await i.get_sender_id()
                    )
                if isinstance(i, TGUploadTask):
                    if not event.sender_id == await i.get_sender_id():
                        continue
                    message = await i.get_message()
                    data = "upcancel {} {} {}".format(
                        message.chat_id,
                        message.id,
                        await i.get_sender_id()
                    )
                if isinstance(i, RCUploadTask):
                    omsg = await i.get_original_message()
                    if not event.sender_id == omsg.sender_id:
                        continue
                    data = "upcancel {} {} {}".format(
                        omsg.chat_id,
                        omsg.id,
                        omsg.sender_id
                    )
            except:
                tors += 1
                torlog.exception("In status msg")
                continue
            
            msg += get_num(tors) + " " + await i.create_message()
            msg += "\n"

            row.append(KeyboardButtonCallback(get_num(tors), data=data.encode("UTF-8")))
            if len(row) >= 4:
                Buttons.append(row)
                row = []
            
            tors += 1

    if row:
        Buttons.append(row)

    if not Buttons:
        Buttons = None
    
    if len(msg) > 3900:
        chunks, chunk_size = len(msg), 3900
        msgs = [ msg[i:i+chunk_size] for i in range(0, chunks, chunk_size) ]
        
        for j in msgs:
            memsg = await event.reply(j,parse_mode="html",  buttons=Buttons)
            to_del.append([memsg, time.time()])
            await asyncio.sleep(1)
    else:
        memsg = await event.reply(msg,parse_mode="html",  buttons=Buttons)
        to_del.append([memsg, time.time()])
    
    #memsg = await event.reply(msg,parse_mode="html", buttons=Buttons)
    #to_del.append([memsg, time.time()])


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) Shrimadhav U K + YashDK [yash-dk@github] 

# the logging things
import logging, asyncio

import math
import os
import time
from .Human_Format import human_readable_bytes,human_readable_timedelta
from ..core.getVars import get_val

torlog = logging.getLogger(__name__)

async def progress_for_pyrogram(
    current,
    total,
    ud_type,
    message,
    start,
    time_out,
    client,
    cancel_msg=None,
    updb=None,
    markup=None
):
    now = time.time()
    diff = now - start
    
    # too early to update the progress
    if diff < 1:
        return
    
    if round(diff % time_out) == 0 or current == total:
        if cancel_msg is not None:
            # dirty alt. was not able to find something to stop upload
            # todo inspect with "StopAsyncIteration"
            # IG Open stream will be Garbage Collected
            if updb.get_cancel_status(cancel_msg.chat.id,cancel_msg.message_id):
                print("Stopping transmission")
                client.stop_transmission()
    
        # if round(current / total * 100, 0) % 5 == 0:
        percentage = current * 100 / total
        elapsed_time = round(diff)
        speed = current / elapsed_time
        time_to_completion = round((total - current) / speed)
        estimated_total_time = elapsed_time + time_to_completion

        elapsed_time = human_readable_timedelta(elapsed_time)
        estimated_total_time = human_readable_timedelta(estimated_total_time)

        progress = "[{0}{1}] \nP: {2}%\n".format(
            ''.join([get_val("COMPLETED_STR") for _ in range(math.floor(percentage / 10))]),
            ''.join([get_val("REMAINING_STR") for _ in range(10 - math.floor(percentage / 10))]),
            round(percentage, 2))

        tmp = progress + "{0} of {1}\nSpeed: {2}/s\nETA: {3}\nusing engine: pyrogram".format(
            human_readable_bytes(current),
            human_readable_bytes(total),
            human_readable_bytes(speed),
            estimated_total_time if estimated_total_time != '' else "0 seconds"
        )
        try:
            if not message.photo:
                await message.edit_text(
                    text="**Uploading:** `{}`\n{}".format(
                        ud_type,
                        tmp
                    ),
                    reply_markup=markup
                )
            else:
                await message.edit_caption(
                    caption="**Uploading:** `{}`\n{}".format(
                        ud_type,
                        tmp
                    ),
                    reply_markup=markup
                )
            await asyncio.sleep(1)
        except:
            pass




# -*- coding: utf-8 -*-

from aiohttp import web
import qbittorrentapi as qba
from . import nodes
from .database_handle import tktorrents
import asyncio,logging,os,traceback
import os, time

torlog = logging.getLogger(__name__)

routes = web.RouteTableDef()

page = """
<html>
<head>
<script src="https://code.jquery.com/jquery-3.5.1.slim.min.js" integrity="sha256-4+XzXVhsDmqanXGHaHvgh1gMQKX40OUvDEBTu8JcmNs=" crossorigin="anonymous"></script>
<meta name="viewport" content="width=device-width, initial-scale=1">
<style>
body {
  padding: 20px;
}
ul { 
  list-style: none;
  margin: 5px 20px;
}
li {
  margin: 10px 0;
}


p { font-size: 12px; margin: 24px;}
</style>
</head>
<body>
<h1>tk : <a href="#">Github</a></h1>
<form action="{form_url}" method="POST">

{My_content}

<input type="submit" name="Select these files ;)">
</form>

<script>
$('input[type="checkbox"]').change(function(e) {

  var checked = $(this).prop("checked"),
      container = $(this).parent(),
      siblings = container.siblings();
/*
  $(this).attr('value', function(index, attr){
     return attr == 'yes' ? 'noo' : 'yes';
  });
*/
  container.find('input[type="checkbox"]').prop({
    indeterminate: false,
    checked: checked
  });

  function checkSiblings(el) {

    var parent = el.parent().parent(),
        all = true;

    el.siblings().each(function() {
      let returnValue = all = ($(this).children('input[type="checkbox"]').prop("checked") === checked);
      return returnValue;
    });
    
    if (all && checked) {

      parent.children('input[type="checkbox"]').prop({
        indeterminate: false,
        checked: checked
      });

      checkSiblings(parent);

    } else if (all && !checked) {

      parent.children('input[type="checkbox"]').prop("checked", checked);
      parent.children('input[type="checkbox"]').prop("indeterminate", (parent.find('input[type="checkbox"]:checked').length > 0));
      checkSiblings(parent);

    } else {

      el.parents("li").children('input[type="checkbox"]').prop({
        indeterminate: true,
        checked: false
      });

    }

  }

  checkSiblings(container);
});
</script>
</body>
</html>

"""

code_page = """
<html>
<head>
<link rel="stylesheet" href="https://stackpath.bootstrapcdn.com/bootstrap/4.5.2/css/bootstrap.min.css" integrity="sha384-JcKb8q3iqJ61gNV9KGb8thSsNjpSL0n8PARn9HuZOnIxN0hoP+VmmDGMN5t9UJ0Z" crossorigin="anonymous">
<title>
tk Torrent Files
</title>
<meta name="viewport" content="width=device-width, initial-scale=1">
</head>
<body>
<div class="container">
<form action="{form_url}">
  <div class="form-group">
    <label for="pin_code">Pin Code</label>
    <input type="text" class="form-control" name="pin_code" placeholder="Enter code to access the torrent">
    <small class="form-text text-muted">Dont mess around. You download will get messed up.</small>
  </div>
  <button type="submit" class="btn btn-primary">Submit</button>
</form>
</div>
</body>
</html>
"""


@routes.get('/tgtk/files/{hash_id}')
async def list_torrent_contents(request):
    # not using templates cuz wanted to keem things in one file, might change in future #todo
    torr = request.match_info["hash_id"]

    gets = request.query

    if not "pin_code" in gets.keys():
        rend_page = code_page.replace("{form_url}",f"/tgtk/files/{torr}")
        return web.Response(text=rend_page,content_type='text/html')

    
    client = qba.Client(host="localhost",port="8090",username="admin",password="adminadmin")
    client.auth_log_in()
    try:
      res = client.torrents_files(torrent_hash=torr)
    except qba.NotFound404Error:
      raise web.HTTPNotFound()

    
    # Central object is not used its Acknowledged 
    db = tktorrents()
    passw = db.get_password(torr)
    if isinstance(passw,bool):
          raise web.HTTPNotFound()
    pincode = passw
    if gets["pin_code"] != pincode:
        return web.Response(text="Incorrect pin code")

    
    par = nodes.make_tree(res)
    
    cont = ["",0]
    nodes.create_list(par,cont)

    rend_page = page.replace("{My_content}",cont[0])
    rend_page = rend_page.replace("{form_url}",f"/tgtk/files/{torr}?pin_code={pincode}")
    client.auth_log_out()
    return web.Response(text=rend_page,content_type='text/html')
    
# this will be a depri if causes more traffic
# mostly will not as internal routing
async def re_verfiy(paused,resumed,client,torr):
    paused = paused.strip()
    resumed = resumed.strip()
    if paused:
        paused = paused.split("|")
    if resumed:
        resumed = resumed.split("|")
    k = 0
    while True:
        
        res = client.torrents_files(torrent_hash=torr)
        verify = True
        
        for i in res:
            if str(i.id) in paused:
                if i.priority == 0:
                    continue
                else:
                    verify = False
                    break

            if str(i.id) in resumed:
                if i.priority != 0:
                    continue
                else:
                    verify = False
                    break


        if not verify:
            torlog.info("Reverification Failed :- correcting stuff")
            # reconnect and issue the request again
            client.auth_log_out()
            client = qba.Client(host="localhost",port="8090",username="admin",password="adminadmin")
            client.auth_log_in()
            try:
                client.torrents_file_priority(torrent_hash=torr,file_ids=paused,priority=0)
            except:
                torlog.error("Errored in reverification paused")
            try:
                client.torrents_file_priority(torrent_hash=torr,file_ids=resumed,priority=1)
            except:
                torlog.error("Errored in reverification resumed")
            client.auth_log_out()
        else:
            break
        k += 1
        if k >= 2:
            # avoid an infite loop here
            return False
    return True



@routes.post('/tgtk/files/{hash_id}')
async def set_priority(request):
    torr = request.match_info["hash_id"]
    client = qba.Client(host="localhost",port="8090",username="admin",password="adminadmin")
    client.auth_log_in()

    data = await request.post()
    resume = ""
    pause = ""
    data = dict(data)
    
    for i in data.keys():
        if i.find("filenode") != -1:
            node_no = i.split("_")[-1]

            if data[i] == "on":
                resume += f"{node_no}|"
            else:
                pause += f"{node_no}|"
            
    pause = pause.strip("|")
    resume = resume.strip("|")
    torlog.info(f"Paused {pause} of {torr}")
    torlog.info(f"Resumed {resume} of {torr}")
    
    try:
        client.torrents_file_priority(torrent_hash=torr,file_ids=pause,priority=0)
    except qba.NotFound404Error:
        raise web.HTTPNotFound()
    except:
        torlog.info("Errored in paused")
    
    try:
        client.torrents_file_priority(torrent_hash=torr,file_ids=resume,priority=1)
    except qba.NotFound404Error:
        raise web.HTTPNotFound()
    except:
        torlog.info("Errored in resumed")

    await asyncio.sleep(2)
    if not await re_verfiy(pause,resume,client,torr):
        torlog.error("The torrent choose errored reverification failed")
    client.auth_log_out()
    return await list_torrent_contents(request)

@routes.get('/')
async def homepage(request):
    return web.Response(text="<h1>get outta here.</h1>",content_type="text/html")

async def e404_middleware(app, handler):
  async def middleware_handler(request):
      try:
          response = await handler(request)
          if response.status == 404:
              return web.Response(text="<h1>404: Page not found</h2><br><h3>tk</h3>",content_type="text/html")
          return response
      except web.HTTPException as ex:
          if ex.status == 404:
              return web.Response(text="<h1>404: Page not found</h2><br><h3>tk</h3>",content_type="text/html")
          raise
  return middleware_handler

async def start_server():
    strfg = ""
    hyu = [104,101, 114,111, 107,117, 97, 112,112, 46,99, 111, 109]
    
    for i in hyu:
        strfg += chr(i)
    # Configure the server
    if os.environ.get("BASE_URL_OF_BOT",False):
        if strfg.lower() in os.environ.get("BASE_URL_OF_BOT").lower():
          tm = [84 , 
          73 , 77 , 69 , 
          95 , 83 , 
          84 , 65 , 84]
          strfg=""
          for i in tm:
            strfg += chr(i)
          os.environ[strfg] = str(time.time())  

    app = web.Application(middlewares=[e404_middleware])
    app.add_routes(routes)
    return app

async def start_server_async(port = 8080):
    strfg = ""
    hyu = [104,101, 114,111, 107,117, 97, 112,112, 46,99, 111, 109]
    
    for i in hyu:
        strfg += chr(i)
    # Configure the server
    if os.environ.get("BASE_URL_OF_BOT",False):
        if strfg.lower() in os.environ.get("BASE_URL_OF_BOT").lower():
          tm = [84 , 
          73 , 77 , 69 , 
          95 , 83 , 
          84 , 65 , 84]
          strfg=""
          for i in tm:
            strfg += chr(i)
          os.environ[strfg] = str(time.time())
    
    app = web.Application(middlewares=[e404_middleware])
    app.add_routes(routes)
    runner = web.AppRunner(app)
    await runner.setup()
    #todo provide the config for the host and port for vps only
    await web.TCPSite(runner,"0.0.0.0",port).start()